/******************************************************************************
 *
 * Source file for harness-dummy (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harness-dummy.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Harness-dummy plug-in module data type. */
typedef struct {
  pthread_mutex_t  mutex;     /* mutex                 */
  unsigned int     count;     /* instances array count */
  struct {
    unsigned int   handle;    /* instance handle       */
  }               *instances; /* instances array       */
} harness_dummy_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the harness-dummy plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_INIT(harness_dummy_init);

/*
 * Finalizes the harness-dummy plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
HARNESS_PLUGINS_INIT(harness_dummy_fini);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Harness-dummy plug-in module data. */
harness_dummy_data_t harness_dummy_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex                 */
  0,                         /* instances array count */
  NULL                       /* instances array       */
};

/* Harness-dummy plug-in "object". */
harness_dummy_t harness_dummy = {
  {
    LIBHARNESS_DUMMY_VERSION_CURRENT,  /* current version         */
    LIBHARNESS_DUMMY_VERSION_REVISION, /* current revision        */
    LIBHARNESS_DUMMY_VERSION_AGE,      /* version age             */
    LIBHARNESS_DUMMY_VERSION_FIRST     /* first supported version */
  },                                   /* plug-in version         */
  /**********************************************/
  /* PUT YOUR PUBLIC PLUG-IN DATA POINTERS HERE */
  /**********************************************/
  /**************************************************/
  /* PUT YOUR PUBLIC PLUG-IN FUNCTION POINTERS HERE */
  /**************************************************/
  harness_dummy_instances
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the harness-dummy plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_dummy_init"
HARNESS_PLUGINS_INIT(harness_dummy_init) {
  int           error;
  void         *instances;
  unsigned int  index;
  /* Lock harness-dummy plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_dummy_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to lock harness-dummy plug-in mutex)))
    harness_syserr();
    return -1;
  }
  /* Search for handle in instances array. */
  for (index = 0; index < harness_dummy_data.count; index++) {
    if (harness_dummy_data.instances[index].handle == handle) {
      HARNESS_LOG((
        HARNESS_WARN(handle is already in instances array)))
      harness_syserr();
      /* Unlock harness-dummy plug-in mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_dummy_data.mutex))) {
        errno = error;
        HARNESS_LOG((
          HARNESS_WARN(unable to unlock harness-dummy plug-in mutex)))
        harness_syserr();
      }
      return -1;
    }
  }
  /* Increase instances array. */
  index = harness_dummy_data.count;
  harness_dummy_data.count++;
  /* Reallocate instances array. */
  if (NULL == (instances = realloc(harness_dummy_data.instances,
                                   harness_dummy_data.count *
                            sizeof(harness_dummy_data.instances[0])))) {
    HARNESS_LOG((
      HARNESS_WARN(unable to reallocate instances array)))
    /* Unlock harness-dummy plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_dummy_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to unlock harness-dummy plug-in mutex)))
      harness_syserr();
    }
    return -1;
  }
  harness_dummy_data.instances = instances;
  /* Save instances array entry. */
  harness_dummy_data.instances[index].handle = handle;
  /* Check for first initialization. */
  if (1 == harness_dummy_data.count) {
    /***************************/
    /* PUT YOUR INIT CODE HERE */
    /***************************/
  }
  /* Unlock harness-dummy plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_dummy_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to unlock harness-dummy plug-in mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Finalizes the harness-dummy plug-in.
 *
 * handle = The plug-in instance handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_dummy_fini"
HARNESS_PLUGINS_INIT(harness_dummy_fini) {
  int           error;
  void         *instances;
  unsigned int  index;
  /* Lock harness-dummy plug-in mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_dummy_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to lock harness-dummy plug-in mutex)))
    harness_syserr();
    return -1;
  }
  /* Search for handle in instances array. */
  for (index = 0; index < harness_dummy_data.count; index++) {
    if (harness_dummy_data.instances[index].handle == handle) {
      break;
    }
  }
  /* Check if handle is not in instances array. */
  if (index == harness_dummy_data.count) {
    HARNESS_LOG((
      HARNESS_WARN(handle is not in instances array)))
    harness_syserr();
    /* Unlock harness-dummy plug-in mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_dummy_data.mutex))) {
      errno = error;
      HARNESS_LOG((
        HARNESS_WARN(unable to unlock harness-dummy plug-in mutex)))
      harness_syserr();
    }
    return -1;
  }
  /* Remove instance from instances array. */
  harness_dummy_data.count--;
  memmove(harness_dummy_data.instances + index,
          harness_dummy_data.instances + index + 1,
         (harness_dummy_data.count     - index) *
          sizeof(unsigned int));
  /* Reallocate instances array. */
  if (0 == harness_dummy_data.count) {
    free(harness_dummy_data.instances);
    harness_dummy_data.instances = NULL;
  } else if (NULL ==
             (instances = realloc(harness_dummy_data.instances,
                                  harness_dummy_data.count *
                           sizeof(harness_dummy_data.instances[0])))) {
    HARNESS_LOG((
      HARNESS_WARN(unable to reallocate instances array)))
  } else {
    harness_dummy_data.instances = instances;
  }
  /* Check for last finalization. */
  if (0 == harness_dummy_data.count) {
    /***************************/
    /* PUT YOUR FINI CODE HERE */
    /***************************/
  }
  /* Unlock harness-dummy plug-in mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_dummy_data.mutex))) {
    errno = error;
    HARNESS_LOG((
      HARNESS_WARN(unable to unlock harness-dummy plug-in mutex)))
    harness_syserr();
    return -1;
  }
  return 0;
}

/*
 * Returns the number of loaded instances of this harness-dummy plug-in without
 * locking its plug-in mutex.
 *
 * return = The number of loaded instances of this harness-dummy plug-in.
 */
#undef  __FUNC__
#define __FUNC__ "harness_dummy_instances"
unsigned int harness_dummy_instances () {
  return harness_dummy_data.count;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
