/******************************************************************************
 *
 * Source file for harness (main) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include <harness.h>

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Harnessd module data type. */
typedef struct {
  unsigned int   daemon;      /* daemon flag           */
  unsigned int   forker;      /* forker flag           */
  void         (*onsyserr)(); /* system error callback */
} harness_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Sets the forker process configuration and runs the forker process.
 *
 * size   = The forker process buffer size.
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_forker_main (unsigned int size,
                         int          pipes[2]);

/*
 * Runs the forker process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_forker_main2 (int pipes[2]);

/*
 * Sets the kernel process configuration and runs the kernel process.
 *
 * pool_minimum = The new minimum thread pool thread count.
 * pool_maximum = The new maximum thread pool thread count.
 * pool_timeout = The new exiting thread pool thread timeout.
 * jobs_maximum = The new maximum thread pool job count.
 * jobs_timeout = The new waiting thread pool thread timeout.
 * init_options = The new init options to be applied at startup.
 *                A NULL terminated array of strings containing option/value
 *                pairs. Possible option/value pairs are:
 *                'load' and 'VALUE' to load a plug-in with the name of VALUE at
 *                    start-up. Allows plug-in to load more than once and does 
 *                    not export its symbols. VALUE cannot be ommitted.
 *                'load-unique' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    does not export its symbols. VALUE cannot be ommitted.
 *                'load-export' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    exports its symbols. VALUE cannot be ommitted.
 *                'exec' and 'VALUE' to execute a program with the name of VALUE
 *                    in a child process at start-up. VALUE cannot be ommitted.
 *                All options are applied in the order they appear.
 * pipes        = The pipe ends for kernel-forker communication.
 * block        = The block flag to indicate wether to run the kernel in this
 *                thread and block until finished, or to create a new kernel
 *                thread and return after kernel startup.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_main (unsigned int   pool_minimum,
                         unsigned int   pool_maximum,
                         unsigned int   pool_timeout,
                         unsigned int   jobs_maximum,
                         unsigned int   jobs_timeout,
                         char         **init_options,
                         int            pipes[2],
                         unsigned int   block);

/*
 * Runs the kernel process.
 *
 * pipes  = The pipe ends for kernel-forker communication.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_main2 (int pipes[2]);

/*
 * Waits for the kernel thread to shut down if started in non-blocking mode.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_kernel_join ();

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Harnessd module data. */
harness_data_t harness_data = {
  1,               /* daemon flag           */
  1,               /* forker flag           */
  HARNESS_ONSYSERR /* system error callback */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Sets the daemon flag.
 * This function may only be called before initializing.
 *
 * daemon = The daemon flag.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_daemon_set"
int harness_daemon_set (unsigned int daemon) {
  /* Check daemon parameter. */
  if (1 < daemon) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(daemon parameter is greater than 1)))
    return -1;
  }
  harness_data.daemon = daemon;
  return 0;
}

/*
 * Sets the forker flag.
 * This function may only be called before initializing.
 *
 * forker = The forker flag.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_set"
int harness_forker_set (unsigned int forker) {
  /* Check forker parameter. */
  if (1 < forker) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(forker parameter is greater than 1)))
    return -1;
  }
  harness_data.forker = forker;
  return 0;
}

/*
 * Sets the callback for reporting system errors.
 *
 * onsyserr = The callback for reporting system errors.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_onsyserr_set"
int harness_onsyserr_set (void (*onsyserr)()) {
  harness_data.onsyserr = onsyserr;
  return 0;
}

/*
 * Gets the daemon flag.
 *
 * daemon = The daemon flag (return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_daemon_get"
int harness_daemon_get (unsigned int *daemon) {
#ifdef DEBUG
  /* Check daemon parameter. */
  if (NULL == daemon) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(daemon parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  *daemon = harness_data.daemon;
  return 0;
}

/*
 * Gets the forker flag.
 *
 * forker = The forker flag (return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_forker_get"
int harness_forker_get (unsigned int *forker) {
#ifdef DEBUG
  /* Check forker parameter. */
  if (NULL == forker) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(forker parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  *forker = harness_data.forker;
  return 0;
}

/*
 * Gets the callback for reporting system errors.
 *
 * onsyserr = The callback for reporting system errors (return).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_onsyserr_get"
int harness_onsyserr_get (void (**onsyserr)()) {
#ifdef DEBUG
  /* Check onsyserr parameter. */
  if (NULL == onsyserr) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(onsyserr parameter is null)))
    return -1;
  }
#endif /* DEBUG */
  *onsyserr = harness_data.onsyserr;
  return 0;
}

/*
 * Sets the kernel configuration and runs the kernel.
 *
 * daemon       = The new daemon flag.
 * forker       = The new forker flag.
 * forker_size  = The forker process buffer size.
 * pool_minimum = The new minimum kernel process thread pool thread count.
 * pool_maximum = The new maximum kernel process thread pool thread count.
 * pool_timeout = The new exiting kernel process thread pool thread timeout.
 * jobs_maximum = The new maximum kernel process thread pool job count.
 * jobs_timeout = The new waiting kernel process thread pool thread timeout.
 * init_options = The new init options to be applied at startup.
 *                A NULL terminated array of strings containing option/value
 *                pairs. Possible option/value pairs are:
 *                'load' and 'VALUE' to load a plug-in with the name of VALUE at
 *                    start-up. Allows plug-in to load more than once and does 
 *                    not export its symbols. VALUE cannot be ommitted.
 *                'load-unique' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    does not export its symbols. VALUE cannot be ommitted.
 *                'load-export' and 'VALUE' to load a plug-in with the name of
 *                    VALUE at start-up. Allows plug-in to load only once and
 *                    exports its symbols. VALUE cannot be ommitted.
 *                'exec' and 'VALUE' to execute a program with the name of VALUE
 *                    in a child process at start-up. VALUE cannot be ommitted.
 *                All options are applied in the order they appear.
 * block        = The block flag to indicate wether to run the kernel in this
 *                thread and block until finished, or to create a new kernel
 *                thread and return after kernel startup.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_main"
int harness_main (unsigned int   daemon,
                  unsigned int   forker,
                  unsigned int   forker_size,
                  unsigned int   pool_minimum,
                  unsigned int   pool_maximum,
                  unsigned int   pool_timeout,
                  unsigned int   jobs_maximum,
                  unsigned int   jobs_timeout,
                  char         **init_options,
                  unsigned int   block) {
  pid_t pid;
  int   pipes[2][2];
  /* Set the daemon flag. */
  if (0 != harness_daemon_set(daemon)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the daemon flag)))
    return -1;
  }
  /* Set the forker flag. */
  if (0 != harness_forker_set(forker)) {
    HARNESS_LOG((HARNESS_WARN(unable to set the forker flag)))
    return -1;
  }
  if (1 == harness_data.daemon) {
    /* Create daemon process. */
    pid = fork();
  } else {
    /* Do not create daemon process. */
    pid = 0;
  }
  switch (pid) {
    /* Check for error. */
    case -1: {
      HARNESS_LOG((HARNESS_WARN(unable to create daemon process)))
      return -1;
    }
    /* Check for child process or if no daemon is created. */
    case 0: {
      /* Check forker flag. */
      if (0 == harness_data.forker) {
        /* Reset pipes. */
        pipes[0][0] = -1;
        pipes[0][1] = -1;
        /* Run kernel process. */
        if (0 != harness_kernel_main(pool_minimum, pool_maximum, pool_timeout,  
                                     jobs_maximum, jobs_timeout, init_options,
                                     pipes[0], block)) {
          HARNESS_LOG((HARNESS_WARN(unable to run kernel process)))
          return -1;
        }
        return 0;
      } else {
        /* Create pipe one. */
        if (0 != pipe(pipes[0])) {
          HARNESS_LOG((HARNESS_WARN(unable to create pipe one)))
          return -1;
        }
        /* Create pipe two. */
        if (0 != pipe(pipes[1])) {
          HARNESS_LOG((HARNESS_WARN(unable to create pipe two)))
          /* Close pipe one. */
          close(pipes[0][0]);
          close(pipes[0][1]);
          return -1;
        }
        /* Create forker process. */
        switch (fork()) {
          /* Check for error. */
          case -1: {
            HARNESS_LOG((HARNESS_WARN(unable to create forker process)))
            /* Close pipes one and two. */
            close(pipes[0][0]);
            close(pipes[0][1]);
            close(pipes[1][0]);
            close(pipes[1][1]);
            return -1;
          }
          /* Check for child process. */
          case 0: {
            /* Close unused pipe ends. */
            close(pipes[0][0]);
            close(pipes[1][1]);
            pipes[0][0] = pipes[1][0];
            /* Run forker process. */
            if (0 != harness_forker_main(forker_size, pipes[0])) {
              HARNESS_LOG((HARNESS_WARN(unable to run forker process)))
              /* Close other pipe ends. */
              close(pipes[0][0]);
              close(pipes[0][1]);
              return -1;
            }
            /* Close other pipe ends. */
            close(pipes[0][0]);
            close(pipes[0][1]);
            return 0;
          }
          /* Check for parent process. */
          default: {
            /* Close unused pipe ends. */
            close(pipes[0][1]);
            close(pipes[1][0]);
            pipes[1][0] = pipes[0][0];
            /* Run kernel process. */
            if (0 != harness_kernel_main(pool_minimum, pool_maximum,
                                         pool_timeout, jobs_maximum,
                                         jobs_timeout, init_options,
                                         pipes[1], block)) {
              HARNESS_LOG((HARNESS_WARN(unable to run kernel process)))
              /* Close other pipe ends. */
              close(pipes[1][0]);
              close(pipes[1][1]);
              return -1;
            }
            return 0;
          }
        }
      }
    }
    /* Check for parent process. */
    default: {
      return 0;
    }
  }
}

/*
 * Runs the kernel.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_main2"
int harness_main2 () {
  pid_t pid;
  int   pipes[2][2];
  if (1 == harness_data.daemon) {
    /* Create daemon process. */
    pid = fork();
  } else {
    /* Do not create daemon process. */
    pid = 0;
  }
  switch (pid) {
    /* Check for error. */
    case -1: {
      HARNESS_LOG((HARNESS_WARN(unable to create daemon process)))
      return -1;
    }
    /* Check for child process or if no daemon is created. */
    case 0: {
      /* Check forker flag. */
      if (0 == harness_data.forker) {
        /* Reset pipes. */
        pipes[0][0] = -1;
        pipes[0][1] = -1;
        /* Run kernel process. */
        if (0 != harness_kernel_main2(pipes[0])) {
          HARNESS_LOG((HARNESS_WARN(unable to run kernel process)))
          return -1;
        }
        return 0;
      } else {
        /* Create pipe one. */
        if (0 != pipe(pipes[0])) {
          HARNESS_LOG((HARNESS_WARN(unable to create pipe one)))
          return -1;
        }
        /* Create pipe two. */
        if (0 != pipe(pipes[1])) {
          HARNESS_LOG((HARNESS_WARN(unable to create pipe two)))
          /* Close pipe one. */
          close(pipes[0][0]);
          close(pipes[0][1]);
          return -1;
        }
        /* Create forker process. */
        switch (fork()) {
          /* Check for error. */
          case -1: {
            HARNESS_LOG((HARNESS_WARN(unable to create forker process)))
            /* Close pipes one and two. */
            close(pipes[0][0]);
            close(pipes[0][1]);
            close(pipes[1][0]);
            close(pipes[1][1]);
            return -1;
          }
          /* Check for child process. */
          case 0: {
            /* Close unused pipe ends. */
            close(pipes[0][0]);
            close(pipes[1][1]);
            pipes[0][0] = pipes[1][0];
            /* Run forker process. */
            if (0 != harness_forker_main2(pipes[0])) {
              HARNESS_LOG((HARNESS_WARN(unable to run forker process)))
              /* Close other pipe ends. */
              close(pipes[0][0]);
              close(pipes[0][1]);
              return -1;
            }
            /* Close other pipe ends. */
            close(pipes[0][0]);
            close(pipes[0][1]);
            return 0;
          }
          /* Check for parent process. */
          default: {
            /* Close unused pipe ends. */
            close(pipes[0][1]);
            close(pipes[1][0]);
            pipes[1][0] = pipes[0][0];
            /* Run kernel process. */
            if (0 != harness_kernel_main2(pipes[1])) {
              HARNESS_LOG((HARNESS_WARN(unable to run kernel process)))
              /* Close other pipe ends. */
              close(pipes[1][0]);
              close(pipes[1][1]);
              return -1;
            }
            return 0;
          }
        }
      }
    }
    /* Check for parent process. */
    default: {
      return 0;
    }
  }
}

/*
 * Reports a system error via callback.
 */
#undef  __FUNC__
#define __FUNC__ "harness_syserr"
void harness_syserr () {
  void (*onsyserr)();
  /* Check for system error callback. */
  onsyserr = harness_data.onsyserr;
  if (NULL != onsyserr) {
    /* Execute system error callback. */
    HARNESS_LOG((HARNESS_INFO(reporting system error via callback)))
    onsyserr();
  } else {
    HARNESS_LOG((HARNESS_INFO(not reporting system error via callback)))
  }
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
