/******************************************************************************
 *
 * Source file for harness plugins (plug-in loader) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "harness.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Plug-in loader data type. */
typedef struct {
  pthread_mutex_t        mutex;     /* mutex                        */
  unsigned int           ready;     /* initialized flag             */
  unsigned int           closing;   /* closing flag                 */
  unsigned int           count;     /* plug-in registry count       */
  struct {
    char                *path;      /* plug-in path                 */
    char                *name;      /* plug-in name                 */
    void                *module;    /* plug-in module               */
    harness_plugins_init_t  init;      /* plug-in init symbol          */
    harness_plugins_fini_t  fini;      /* plug-in fini symbol          */
    harness_plugins_init_t  init2;     /* plug-in init symbol (global) */
    harness_plugins_fini_t  fini2;     /* plug-in fini symbol (global) */
    unsigned int         unique;    /* plug-in unique flag          */
    unsigned int         count;     /* plug-in instance array count */
    struct {
      unsigned int       handle;    /* plug-in instance handle      */
      unsigned int       ready;     /* plug-in instance ready flag  */
    }                   *instances; /* plug-in instance array       */
  }                     *plugins;   /* plug-in registry array       */
} harness_plugins_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the plug-in loader.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_plugins_init ();

/*
 * Finalizes the plug-in loader.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int harness_plugins_fini ();

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Plug-in loader data. */
harness_plugins_data_t harness_plugins_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex                  */
  0,                         /* initialized flag       */
  0,                         /* closing flag           */
  0,                         /* plug-in registry count */
  NULL                       /* plug-in registry array */
};

/* Plug-in loader directory. */
const char *harness_plugins_dir = HARNESS_STRING(HARNESS_LOADDIR);

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the plug-in loader.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_init"
int harness_plugins_init () {
  int error;
  HARNESS_LOG((HARNESS_INFO(plug-in loader startup)))
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check for initialized plug-in loader. */
  if (1 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader already initialized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Initialize plug-in registry. */
  harness_plugins_data.count   = 0;
  harness_plugins_data.plugins = NULL;
  /* Mark plug-in loader initialized. */
  harness_plugins_data.closing = 0;
  harness_plugins_data.ready   = 1;
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    /* Mark plug-in loader finalized. */
    harness_plugins_data.ready = 0;
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the plug-in loader.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_fini"
int harness_plugins_fini () {
  int                  error, result, errno2;
  void                *module, *instances, *plugins;
  unsigned int         handle;
  harness_plugins_fini_t  fini;
  HARNESS_LOG((HARNESS_INFO(plug-in loader shutdown)))
  result = 0;
  errno2 = 0;
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader already finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Mark plug-in loader closing. */
  harness_plugins_data.closing = 1;
  /* Unload all plug-ins in registry. */
  while (0 != harness_plugins_data.count) {
    /* Unload first plug-in instance in registry. */
    handle = harness_plugins_data.plugins[0].instances[0].handle;
    /* Check ready flag. */
    if (1 == harness_plugins_data.plugins[0].instances[0].ready) {
      /* Save fini symbol and module. */
      fini   = harness_plugins_data.plugins[0].fini;
      module = harness_plugins_data.plugins[0].module;
    } else {
      /* Save module. */
      fini   = NULL;
      module = harness_plugins_data.plugins[0].module;
    }
    /* Remove instance from instances array. */
    harness_plugins_data.plugins[0].count--;
    memmove(harness_plugins_data.plugins[0].instances,
            harness_plugins_data.plugins[0].instances + 1,
           (harness_plugins_data.plugins[0].count) *
     sizeof(harness_plugins_data.plugins[0].instances[0]));
    /* Reallocate instances array. */
    if (0 == harness_plugins_data.plugins[0].count) {
      free(harness_plugins_data.plugins[0].instances);
      harness_plugins_data.plugins[0].instances = NULL;
    } else  if (NULL == (instances =
                realloc(harness_plugins_data.plugins[0].instances,
                        harness_plugins_data.plugins[0].count *
                 sizeof(harness_plugins_data.plugins[0].instances[0])))) {
      HARNESS_LOG((HARNESS_INFO(unable to reallocate instances array)))
      errno = 0;
    } else {
      harness_plugins_data.plugins[0].instances = instances;
    }
    /* Check if plug-in needs to be removed from plug-in registry. */
    if (0 == harness_plugins_data.plugins[0].count) {
      /* Deallocate path. */
      free(harness_plugins_data.plugins[0].path);
      /* Deallocate name. */
      free(harness_plugins_data.plugins[0].name);
      /* Remove plug-in from plug-in registry. */
      harness_plugins_data.count--;
      memmove(harness_plugins_data.plugins,
              harness_plugins_data.plugins + 1,
             (harness_plugins_data.count) *
       sizeof(harness_plugins_data.plugins[0]));
      /* Reallocate plug-in registry. */
      if (0 == harness_plugins_data.count) {
        free(harness_plugins_data.plugins);
        harness_plugins_data.plugins = NULL;
      } else if (NULL == (plugins =
                          realloc(harness_plugins_data.plugins,
                                  harness_plugins_data.count *
                           sizeof(harness_plugins_data.plugins[0])))) {
        HARNESS_LOG((HARNESS_INFO(unable to reallocate plug-in registry)))
        errno = 0;
      } else {
        harness_plugins_data.plugins = plugins;
      }
    }
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      harness_syserr();
      return -1;
    }
    /* Finalize plug-in. */
    if ((NULL != fini)&&(0 != fini(handle))) {
      errno2 = errno;
      result = -1;
      HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in)))
    }
    /* Unload plug-in. */
    if (0 != (error = dlclose(module))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = error;
      harness_syserr();
      errno = error;
      return -1;
    }
    /* Lock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
      errno = error;
      harness_syserr();
      errno = error;
      return -1;
    }
  }
  /* Mark plug-in loader finalized. */
  harness_plugins_data.closing = 0;
  harness_plugins_data.ready   = 0;
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Loads a plug-in.
 *
 * handle = The plug-in handle (return). Plug-in handles are unique.
 * name   = The plug-in name. Plug-in names are unique and do
 *          not contain '/..' and do not start with '/'.
 * flags  = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_load"
int harness_plugins_load (unsigned int *handle,
                       const char   *name,
                       unsigned int  flags) {
  int                  error;
  void                *module,  *instances, *plugins;
  char                *namedup, *prefix,    *path,   *func, *str;
  unsigned int         namelen,  prefixlen,  pathlen, funclen,
                       index1, index2, index3, index4;
  harness_plugins_init_t  init, init2;
  harness_plugins_fini_t  fini, fini2;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check name parameter. */
  if ('/' == name[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name starts with '/')))
    errno = EINVAL;
    return -1;
  }
  if (NULL != strstr(name, "/..")) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter contains "/..")))
    errno = EINVAL;
    return -1;
  }
  /* Allocate name duplicate. */
  namelen = strlen(name);
  /* Check for .so ending. */
  if ((2 < namelen)&&(0 == strcmp(name + namelen - 3, ".so"))) {
    /* Remove .so ending. */
    namelen -= 3;
  }
  /* Check for .debug ending. */
  if ((5 < namelen)&&(0 == strcmp(name + namelen - 6, ".debug"))) {
    /* Remove .debug ending. */
    namelen -= 6;
  }
  if (NULL == (namedup = (char*)malloc((namelen + 1) * sizeof(char)))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to allocate name duplicate)))
    errno = errno2;
    return -1;
  }
  /* Compose name duplicate. */
  memcpy(namedup, name, namelen);
  namedup[namelen] = '\0';
  /* Search for last digit of version triplet or version number. */
  for (index1 = namelen; 0 < index1-- && '.' != namedup[index1] &&
                                         '0' <= namedup[index1] &&
                                         '9' >= namedup[index1];);
  /* Check for last digit of version triplet or version number. */
  if ((0 != index1)&&('.' == namedup[index1])&&(index1 != namelen - 1)) {
    /* Search for previous digit of version triplet. */
    for (index2 = index1; 0 < index2-- && '.' != namedup[index2] &&
                                          '0' <= namedup[index2] &&
                                          '9' >= namedup[index2];);
    /* Check for previous digit of version triplet. */
    if ((0 != index2)&&('.' == namedup[index2])&&(index2 != index1 - 1)) {
      /* Search for previous digit of version triplet. */
      for (index3 = index2; 0 < index3-- && '.' != namedup[index3] &&
                                            '0' <= namedup[index3] &&
                                            '9' >= namedup[index3];);
      /* Check for previous digit of version triplet. */
      if ((0 != index3)&&('.' == namedup[index3])&&(index3 != index2 - 1)) {
        /* Separate the triplet. */
        namedup[index1] = '\0';
        namedup[index2] = '\0';
        namedup[index3] = '\0';
        /* Load plug-in with version triplet. */
        if (0 != harness_plugins_load3(handle, namedup,
                                    atoi(&namedup[index3 + 1]),
                                    atoi(&namedup[index2 + 1]),
                                    atoi(&namedup[index1 + 1]), flags)) {
          int errno2 = errno;
          HARNESS_LOG((
            HARNESS_WARN(unable to load plug-in with version triplet)))
          /* Deallocate name duplicate. */
          free(namedup);
          errno = errno2;
          return -1;
        }
        /* Deallocate name duplicate. */
        free(namedup);
        return 0;
      }
    } else {
      /* Separate the version number. */
      namedup[index1] = '\0';
      /* Load plug-in with version number. */
      if (0 != harness_plugins_load2(handle, namedup,
                                  atoi(&namedup[index1 + 1]), flags)) {
        int errno2 = errno;
        HARNESS_LOG((
          HARNESS_WARN(unable to load plug-in with version number)))
        /* Deallocate name duplicate. */
        free(namedup);
        errno = errno2;
        return -1;
      }
      /* Deallocate name duplicate. */
      free(namedup);
      return 0;
    }
  }
  /* Allocate init/fini function name. */
  if (NULL == (func = (char*)malloc((namelen + 6) * sizeof(char)))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to allocate name duplicate)))
    errno = errno2;
    return -1;
  }
  /* Compose init/fini function name. */
  memcpy(func, namedup, namelen + 1);
  /* Remove lib prefix from function name. */
  if (NULL == (str = strrchr(func, '/'))) {
    str = func - 1;
  }
  str++;
  if (0 == strncmp(str, "lib", 3)) {
    memmove(str, str + 3, namelen - 2 - (str - func));
    funclen = namelen - 3;
  } else {
    funclen = namelen;
  }
  /* Replace unusual characters in function name. */
  for (index1 = 0; index1 < funclen; index1++) {
    if ((('0' > func[index1])||('9' < func[index1]))&&
        (('a' > func[index1])||('z' < func[index1]))&&
        (('A' > func[index1])||('Z' < func[index1]))) {
      func[index1] = '_';
    }
  }
  /* Allocate path. */
  prefix    =        HARNESS_STRING(HARNESS_LOADDIR);
  prefixlen = strlen(HARNESS_STRING(HARNESS_LOADDIR));
  pathlen   = prefixlen + namelen + 11;
  if (NULL == (path = (char*)malloc(pathlen * sizeof(char)))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to allocate path)))
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = errno2;
    return -1;
  }
  /* Compose path. */
#ifdef    DEBUG
  sprintf(path, "%s/%s.debug.so", prefix, namedup);
#else  /* DEBUG */
  sprintf(path, "%s/%s.so", prefix, namedup);
#endif /* DEBUG */
  /* Load plug-in. */
  if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_LOCAL))) {
    HARNESS_LOG((
      HARNESS_INFO(unable to load %s plug-in: %s), name, dlerror()))
    int errno2 = errno;
    /* Recompose path. */
#ifdef    DEBUG
    sprintf(path, "%s/%s.so", prefix, namedup);
#else  /* DEBUG */
    sprintf(path, "%s/%s.debug.so", prefix, namedup);
#endif /* DEBUG */
    /* Load plug-in. */
    if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_LOCAL))) {
      HARNESS_LOG((
        HARNESS_WARN(unable to load %s plug-in: %s), name, dlerror()))
      /* Deallocate path. */
      free(path);
      /* Deallocate init/fini function name. */
      free(func);
      /* Deallocate name duplicate. */
      free(namedup);
      errno = errno2;
      return -1;
    }
  }
  /* Resolve init function name. */
  memcpy(func + funclen, "_init", 6);
  init  = dlsym(module, func);
  init2 = NULL;
  /* Resolve fini function name. */
  memcpy(func + funclen, "_fini", 6);
  fini  = dlsym(module, func);
  fini2 = NULL;
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    /* Unload plug-in. */
    if (0 != dlclose(module)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = errno2;
      harness_syserr();
    }
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = error;
    return -1;
  }
  /* Check for closing plug-in loader. */
  if (1 == harness_plugins_data.closing) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader is closing)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = EACCES;
      harness_syserr();
    }
    /* Unload plug-in. */
    if (0 != dlclose(module)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = errno2;
      harness_syserr();
    }
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = EACCES;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    /* Unload plug-in. */
    if (0 != dlclose(module)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = errno2;
      harness_syserr();
    }
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = EACCES;
    return -1;
  }
  /* Search for path in plug-in registry. */
  for (index1 = 0; index1 < harness_plugins_data.count; index1++) {
    if (0 == strcmp(harness_plugins_data.plugins[index1].path, path)) {
      break;
    }
  }
  /* Check if path is in plug-in registry. */
  if (index1 < harness_plugins_data.count) {
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Compare names. */
    if (0 != strcmp(harness_plugins_data.plugins[index1].name, namedup)) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded under different name)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      /* Deallocate name duplicate. */
      free(namedup);
      errno = EACCES;
      return -1;
    }
    /* Deallocate name duplicate. */
    free(namedup);
    /* Compare modules. */
    if (harness_plugins_data.plugins[index1].module != module) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded under different module)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Compare init symbols. */
    if (harness_plugins_data.plugins[index1].init != init) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded with different init)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Compare fini symbols. */
    if (harness_plugins_data.plugins[index1].fini != fini) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded with different fini)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Check plug-in unique flag. */
    if ((1 == harness_plugins_data.plugins[index1].unique)||
        ((HARNESS_PLUGINS_UNIQUE == (HARNESS_PLUGINS_UNIQUE & flags)))) {
      errno = EALREADY;
      HARNESS_LOG((HARNESS_WARN(unable to load plug-in more than once)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EALREADY;
      return -1;
    }
    /* Increase instances array. */
    index2 = harness_plugins_data.plugins[index1].count;
    harness_plugins_data.plugins[index1].count++;
    /* Reallocate instances array. */
    if (NULL == (instances =
                 realloc(harness_plugins_data.plugins[index1].instances,
                         harness_plugins_data.plugins[index1].count *
                  sizeof(harness_plugins_data.plugins[0].instances[0])))) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to reallocate instances array)))
      harness_plugins_data.plugins[index1].count--;
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno3 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno3;
        harness_syserr();
      }
      errno = errno2;
      return -1;
    }
    harness_plugins_data.plugins[index1].instances = instances;
  } else {
    /* Check if name appears in plug-in registry. */
    index1 = 0;
    do {
      /* Search for name in plug-in registry. */
      for (; index1 < harness_plugins_data.count; index1++) {
        if (0 == strcmp(harness_plugins_data.plugins[index1].name, name)) {
          break;
        }
      }
      /* Check if name is in plug-in registry. */
      if (index1 < harness_plugins_data.count) {
        /* Check plug-in unique flag. */
        if ((1 == harness_plugins_data.plugins[index1].unique)||
            (HARNESS_PLUGINS_UNIQUE == (HARNESS_PLUGINS_UNIQUE & flags))) {
          errno = EALREADY;
          HARNESS_LOG((HARNESS_WARN(unable to load plug-in more than once)))
          /* Unlock plug-in loader mutex. */
          if (0 != (error =
                    pthread_mutex_unlock(&harness_plugins_data.mutex))) {
            errno = error;
            HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
            errno = error;
            harness_syserr();
          }
          /* Unload plug-in. */
          if (0 != dlclose(module)) {
            int errno2 = errno;
            HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
            errno = errno2;
            harness_syserr();
          }
          /* Deallocate init/fini function name. */
          free(func);
          /* Deallocate path. */
          free(path);
          /* Deallocate name duplicate. */
          free(namedup);
          errno = EALREADY;
          return -1;
        }
        index1++;
      }
    } while (index1 < harness_plugins_data.count);
    /* Check load plug-in globally flag to reload plug-in globally. */
    if (HARNESS_PLUGINS_EXPORT == (HARNESS_PLUGINS_EXPORT & flags)) {
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
        /* Unlock plug-in loader mutex. */
        if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
          errno = error;
          HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
          errno = error;
          harness_syserr();
        }
        /* Deallocate path. */
        free(path);
        /* Deallocate init/fini function name. */
        free(func);
        /* Deallocate name duplicate. */
        free(namedup);
        errno = errno2;
        return -1;
      }
      /* Compose path. */
#ifdef    DEBUG
      sprintf(path, "%s/%s.debug.so", prefix, namedup);
#else  /* DEBUG */
      sprintf(path, "%s/%s.so", prefix, namedup);
#endif /* DEBUG */
      /* Load plug-in. */
      if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_GLOBAL))) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_INFO(unable to reload %s plug-in globally: %s),\
                 name, dlerror()))
        /* Recompose path. */
#ifdef    DEBUG
        sprintf(path, "%s/%s.so", prefix, namedup);
#else  /* DEBUG */
        sprintf(path, "%s/%s.debug.so", prefix, namedup);
#endif /* DEBUG */
        /* Load plug-in. */
        if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_GLOBAL))) {
          HARNESS_LOG((
            HARNESS_WARN(unable to reload %s plug-in globally: %s),\
                   name, dlerror()))
          /* Unlock plug-in loader mutex. */
          if (0 != (error =
                    pthread_mutex_unlock(&harness_plugins_data.mutex))) {
            errno = error;
            HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
            errno = error;
            harness_syserr();
          }
          /* Deallocate path. */
          free(path);
          /* Deallocate init/fini function name. */
          free(func);
          /* Deallocate name duplicate. */
          free(namedup);
          errno = errno2;
          return -1;
        }
      }
      /* Resolve init function name. */
      memcpy(func + namelen, "_init", 6);
      init2 = dlsym(module, func);
      /* Resolve fini function name. */
      memcpy(func + namelen, "_fini", 6);
      fini2 = dlsym(module, func);
      /* Deallocate init/fini function name. */
      free(func);
    } else {
      /* Deallocate init/fini function name. */
      free(func);
    }
    /* Increase plug-in registry. */
    harness_plugins_data.count++;
    /* Reallocate plug-in registry. */
    if (NULL == (plugins = realloc(harness_plugins_data.plugins,
                                   harness_plugins_data.count *
                            sizeof(harness_plugins_data.plugins[0])))) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to reallocate plug-in registry)))
      harness_plugins_data.count--;
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno3 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno3;
        harness_syserr();
      }
      /* Deallocate path. */
      free(path);
      /* Deallocate name duplicate. */
      free(namedup);
      errno = errno2;
      return -1;
    }
    harness_plugins_data.plugins = plugins;
    /* Save plug-in registry entry. */
    harness_plugins_data.plugins[index1].path   = path;
    harness_plugins_data.plugins[index1].name   = namedup;
    harness_plugins_data.plugins[index1].module = module;
    harness_plugins_data.plugins[index1].init   = init;
    harness_plugins_data.plugins[index1].fini   = fini;
    harness_plugins_data.plugins[index1].init2  = init2;
    harness_plugins_data.plugins[index1].fini2  = fini2;
    harness_plugins_data.plugins[index1].unique =
      (HARNESS_PLUGINS_UNIQUE == (HARNESS_PLUGINS_UNIQUE & flags))?1:0;
    /* Increase instances array. */
    index2 = 0;
    harness_plugins_data.plugins[index1].count = 1;
    /* Allocate instances array. */
    if (NULL == (harness_plugins_data.plugins[index1].instances = malloc(
          sizeof(harness_plugins_data.plugins[index1].instances[0])))) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to allocate instances array)))
      /* Decrease plug-in registry. */
      harness_plugins_data.count--;
      /* Reallocate plug-in registry. */
      if (0 == harness_plugins_data.count) {
        free(harness_plugins_data.plugins);
        harness_plugins_data.plugins = NULL;
      } else if (NULL == (plugins =
                          realloc(harness_plugins_data.plugins,
                                  harness_plugins_data.count *
                           sizeof(harness_plugins_data.plugins[0])))) {
        HARNESS_LOG((HARNESS_WARN(unable to reallocate plug-in registry)))
        errno = 0;
      } else {
        harness_plugins_data.plugins = plugins;
      }
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno3 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno3;
        harness_syserr();
      }
      /* Deallocate path. */
      free(path);
      /* Deallocate name duplicate. */
      free(namedup);
      errno = errno2;
      return -1;
    }
  }
  /* Randomly generate a unique non-zero 32-bit handle. */
  harness_plugins_data.plugins[index1].instances[index2].handle = 0;
  do {
    *handle = random() & 0xFFFFFFFF;
    for (index3 = 0;
         index3 < harness_plugins_data.count;
         index3++) {
      for (index4 = 0;
           index4 < harness_plugins_data.plugins[index3].count;
           index4++) {
        if (*handle ==
            harness_plugins_data.plugins[index3].instances[index4].handle) {
          break;
        }
      }
      if (index4 < harness_plugins_data.plugins[index3].count) {
        break;
      }
    }
  } while (index3 < harness_plugins_data.count);
  harness_plugins_data.plugins[index1].instances[index2].handle = *handle;
  /* Check if plug-in needs to be initialized. */
  if (NULL != init) {
    /* Reset ready flag. */
    harness_plugins_data.plugins[index1].instances[index2].ready = 0;
  } else {
    /* Set ready flag. */
    harness_plugins_data.plugins[index1].instances[index2].ready = 1;
  }
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check if plug-in does not need to be initialized. */
  if (NULL == init) {
    HARNESS_LOG((HARNESS_INFO(no init for %s plug-in), name))
    return 0;
  }
  /* Initialize plug-in. */
  if (NULL != init2) {
    if (0 != init2(*handle)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to initialize plug-in)))
      /* Unload plug-in. */
      if (0 != harness_plugins_unload(*handle)) {
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      }
      errno = errno2;
      return -1;
    }
  } else {
    if (0 != init(*handle)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to initialize plug-in)))
      /* Unload plug-in. */
      if (0 != harness_plugins_unload(*handle)) {
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      }
      errno = errno2;
      return -1;
    }
  }
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Search for plug-in in registry. */
  for (index1 = 0, index2 = 0;
       index1 < harness_plugins_data.count;
       index1++) {
    for (index2 = 0;
         index2 < harness_plugins_data.plugins[index1].count;
         index2++) {
      if (*handle ==
          harness_plugins_data.plugins[index1].instances[index2].handle) {
        break;
      }
    }
    if (index2 < harness_plugins_data.plugins[index1].count) {
      break;
    }
  }
  if (index1 == harness_plugins_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(unable to find plug-in in registry)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Set ready flag. */
  harness_plugins_data.plugins[index1].instances[index2].ready = 1;
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Loads a plug-in.
 *
 * handle  = The plug-in handle (return). Plug-in handles are unique.
 * name    = The plug-in name. Plug-in names are unique and do
 *           not contain '/..' and do not start with '/'.
 * version = The version of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_load2"
int harness_plugins_load2 (unsigned int *handle,
                        const char   *name,
                        unsigned int  version,
                        unsigned int  flags) {
  int                  error;
  void                *module,  *instances, *plugins;
  char                *namedup, *prefix,    *path,   *func, *str;
  unsigned int         namelen,  prefixlen,  pathlen, funclen,
                       index1, index2, index3, index4;
  harness_plugins_init_t  init, init2;
  harness_plugins_fini_t  fini, fini2;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check name parameter. */
  if ('/' == name[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name starts with '/')))
    errno = EINVAL;
    return -1;
  }
  if (NULL != strstr(name, "/..")) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter contains "/..")))
    errno = EINVAL;
    return -1;
  }
  /* Allocate name duplicate. */
  namelen = strlen(name);
  /* Check for .so ending. */
  if ((2 < namelen)&&(0 == strcmp(name + namelen - 3, ".so"))) {
    /* Remove .so ending. */
    namelen -= 3;
  }
  /* Check for .debug ending. */
  if ((5 < namelen)&&(0 == strcmp(name + namelen - 6, ".debug"))) {
    /* Remove .debug ending. */
    namelen -= 6;
  }
  if (NULL == (namedup = (char*)malloc((namelen + 1) * sizeof(char)))) {
    HARNESS_LOG((HARNESS_WARN(unable to allocate name duplicate)))
    return -1;
  }
  /* Compose name duplicate. */
  memcpy(namedup, name, namelen);
  namedup[namelen] = '\0';
  /* Allocate init/fini function name. */
  if (NULL == (func = (char*)malloc((namelen + 6) * sizeof(char)))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to allocate name duplicate)))
    errno = errno2;
    return -1;
  }
  /* Compose init/fini function name. */
  memcpy(func, namedup, namelen + 1);
  /* Remove lib prefix from function name. */
  if (NULL == (str = strrchr(func, '/'))) {
    str = func - 1;
  }
  str++;
  if (0 == strncmp(str, "lib", 3)) {
    memmove(str, str + 3, namelen - 2 - (str - func));
    funclen = namelen - 3;
  } else {
    funclen = namelen;
  }
  /* Replace unusual characters in function name. */
  for (index1 = 0; index1 < funclen; index1++) {
    if ((('0' > func[index1])||('9' < func[index1]))&&
        (('a' > func[index1])||('z' < func[index1]))&&
        (('A' > func[index1])||('Z' < func[index1]))) {
      func[index1] = '_';
    }
  }
  /* Allocate path. */
  prefix    =        HARNESS_STRING(HARNESS_LOADDIR);
  prefixlen = strlen(HARNESS_STRING(HARNESS_LOADDIR));
  pathlen   = prefixlen + namelen + 31;
  if (NULL == (path = (char*)malloc(pathlen * sizeof(char)))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to allocate path)))
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = errno2;
    return -1;
  }
  /* Compose path. */
#ifdef    DEBUG
  sprintf(path, "%s/%s.%u.debug.so", prefix, namedup, version);
#else  /* DEBUG */
  sprintf(path, "%s/%s.%u.so", prefix, namedup, version);
#endif /* DEBUG */
  /* Load plug-in. */
  if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_LOCAL))) {
    HARNESS_LOG((
      HARNESS_INFO(unable to load %s plug-in: %s), name, dlerror()))
    int errno2 = errno;
    /* Recompose path. */
#ifdef    DEBUG
    sprintf(path, "%s/%s.%u.so", prefix, namedup, version);
#else  /* DEBUG */
    sprintf(path, "%s/%s.%u.debug.so", prefix, namedup, version);
#endif /* DEBUG */
    /* Load plug-in. */
    if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_LOCAL))) {
      HARNESS_LOG((
        HARNESS_WARN(unable to load %s plug-in: %s), name, dlerror()))
      /* Deallocate path. */
      free(path);
      /* Deallocate init/fini function name. */
      free(func);
      /* Deallocate name duplicate. */
      free(namedup);
      errno = errno2;
      return -1;
    }
  }
  /* Resolve init function name. */
  memcpy(func + funclen, "_init", 6);
  init  = dlsym(module, func);
  init2 = NULL;
  /* Resolve fini function name. */
  memcpy(func + funclen, "_fini", 6);
  fini  = dlsym(module, func);
  fini2 = NULL;
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    harness_syserr();
    /* Unload plug-in. */
    if (0 != dlclose(module)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = errno2;
      harness_syserr();
    }
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = error;
    return -1;
  }
  /* Check for closing plug-in loader. */
  if (1 == harness_plugins_data.closing) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader is closing)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    /* Unload plug-in. */
    if (0 != dlclose(module)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = errno2;
      harness_syserr();
    }
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = EACCES;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    /* Unload plug-in. */
    if (0 != dlclose(module)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = errno2;
      harness_syserr();
    }
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = EACCES;
    return -1;
  }
  /* Search for path in plug-in registry. */
  for (index1 = 0; index1 < harness_plugins_data.count; index1++) {
    if (0 == strcmp(harness_plugins_data.plugins[index1].path, path)) {
      break;
    }
  }
  /* Check if path is in plug-in registry. */
  if (index1 < harness_plugins_data.count) {
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Compare names. */
    if (0 != strcmp(harness_plugins_data.plugins[index1].name, namedup)) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded under different name)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      /* Deallocate name duplicate. */
      free(namedup);
      errno = EACCES;
      return -1;
    }
    /* Deallocate name duplicate. */
    free(namedup);
    /* Compare modules. */
    if (harness_plugins_data.plugins[index1].module != module) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded under different module)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Compare init symbols. */
    if (harness_plugins_data.plugins[index1].init != init) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded with different init)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Compare fini symbols. */
    if (harness_plugins_data.plugins[index1].fini != fini) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded with different fini)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Check plug-in unique flag. */
    if ((1 == harness_plugins_data.plugins[index1].unique)||
        (HARNESS_PLUGINS_UNIQUE == (HARNESS_PLUGINS_UNIQUE & flags))) {
      errno = EALREADY;
      HARNESS_LOG((HARNESS_WARN(unable to load plug-in more than once)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EALREADY;
      return -1;
    }
    /* Increase instances array. */
    index2 = harness_plugins_data.plugins[index1].count;
    harness_plugins_data.plugins[index1].count++;
    /* Reallocate instances array. */
    if (NULL == (instances =
                 realloc(harness_plugins_data.plugins[index1].instances,
                         harness_plugins_data.plugins[index1].count *
                  sizeof(harness_plugins_data.plugins[0].instances[0])))) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to reallocate instances array)))
      harness_plugins_data.plugins[index1].count--;
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno3 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno3;
        harness_syserr();
      }
      errno = errno2;
      return -1;
    }
    harness_plugins_data.plugins[index1].instances = instances;
  } else {
    /* Check if name appears in plug-in registry. */
    index1 = 0;
    do {
      /* Search for name in plug-in registry. */
      for (; index1 < harness_plugins_data.count; index1++) {
        if (0 == strcmp(harness_plugins_data.plugins[index1].name, name)) {
          break;
        }
      }
      /* Check if name is in plug-in registry. */
      if (index1 < harness_plugins_data.count) {
        /* Check plug-in unique flag. */
        if ((1 == harness_plugins_data.plugins[index1].unique)||
            (HARNESS_PLUGINS_UNIQUE == (HARNESS_PLUGINS_UNIQUE & flags))) {
          errno = EALREADY;
          HARNESS_LOG((HARNESS_WARN(unable to load plug-in more than once)))
          /* Unlock plug-in loader mutex. */
          if (0 != (error =
                    pthread_mutex_unlock(&harness_plugins_data.mutex))) {
            errno = error;
            HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
            errno = error;
            harness_syserr();
          }
          /* Unload plug-in. */
          if (0 != dlclose(module)) {
            int errno2 = errno;
            HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
            errno = errno2;
            harness_syserr();
          }
          /* Deallocate init/fini function name. */
          free(func);
          /* Deallocate path. */
          free(path);
          /* Deallocate name duplicate. */
          free(namedup);
          return -1;
        }
        index1++;
      }
    } while (index1 < harness_plugins_data.count);
    /* Check load plug-in globally flag to reload plug-in globally. */
    if ((HARNESS_PLUGINS_EXPORT == (HARNESS_PLUGINS_EXPORT & flags))) {
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        harness_syserr();
        /* Unlock plug-in loader mutex. */
        if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
          errno = error;
          HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
          errno = error;
          harness_syserr();
        }
        /* Deallocate path. */
        free(path);
        /* Deallocate init/fini function name. */
        free(func);
        /* Deallocate name duplicate. */
        free(namedup);
        errno = errno2;
        return -1;
      }
      /* Compose path. */
#ifdef    DEBUG
      sprintf(path, "%s/%s.debug.so", prefix, namedup);
#else  /* DEBUG */
      sprintf(path, "%s/%s.so", prefix, namedup);
#endif /* DEBUG */
      /* Load plug-in. */
      if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_GLOBAL))) {
        HARNESS_LOG((HARNESS_INFO(unable to reload %s plug-in globally: %s),\
                 name, dlerror()))
        int errno2 = errno;
        /* Recompose path. */
#ifdef    DEBUG
        sprintf(path, "%s/%s.so", prefix, namedup);
#else  /* DEBUG */
        sprintf(path, "%s/%s.debug.so", prefix, namedup);
#endif /* DEBUG */
        /* Load plug-in. */
        if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_GLOBAL))) {
          HARNESS_LOG((
            HARNESS_WARN(unable to reload %s plug-in globally: %s),\
                   name, dlerror()))
          /* Unlock plug-in loader mutex. */
          if (0 != (error =
                    pthread_mutex_unlock(&harness_plugins_data.mutex))) {
            errno = error;
            HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
            errno = error;
            harness_syserr();
          }
          /* Deallocate path. */
          free(path);
          /* Deallocate init/fini function name. */
          free(func);
          /* Deallocate name duplicate. */
          free(namedup);
          errno = errno2;
          return -1;
        }
      }
      /* Resolve init function name. */
      memcpy(func + funclen, "_init", 6);
      init2 = dlsym(module, func);
      /* Resolve fini function name. */
      memcpy(func + funclen, "_fini", 6);
      fini2 = dlsym(module, func);
      /* Deallocate init/fini function name. */
      free(func);
    } else {
      /* Deallocate init/fini function name. */
      free(func);
    }
    /* Increase plug-in registry. */
    harness_plugins_data.count++;
    /* Reallocate plug-in registry. */
    if (NULL == (plugins = realloc(harness_plugins_data.plugins,
                                   harness_plugins_data.count *
                            sizeof(harness_plugins_data.plugins[0])))) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to reallocate plug-in registry)))
      harness_plugins_data.count--;
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno3 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno3;
        harness_syserr();
      }
      /* Deallocate path. */
      free(path);
      /* Deallocate name duplicate. */
      free(namedup);
      errno = errno2;
      return -1;
    }
    harness_plugins_data.plugins = plugins;
    /* Save plug-in registry entry. */
    harness_plugins_data.plugins[index1].path   = path;
    harness_plugins_data.plugins[index1].name   = namedup;
    harness_plugins_data.plugins[index1].module = module;
    harness_plugins_data.plugins[index1].init   = init;
    harness_plugins_data.plugins[index1].fini   = fini;
    harness_plugins_data.plugins[index1].init2  = init2;
    harness_plugins_data.plugins[index1].fini2  = fini2;
    harness_plugins_data.plugins[index1].unique =
      (HARNESS_PLUGINS_UNIQUE == (HARNESS_PLUGINS_UNIQUE & flags))?1:0;
    /* Increase instances array. */
    index2 = 0;
    harness_plugins_data.plugins[index1].count = 1;
    /* Allocate instances array. */
    if (NULL == (harness_plugins_data.plugins[index1].instances = malloc(
          sizeof(harness_plugins_data.plugins[index1].instances[0])))) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to allocate instances array)))
      /* Decrease plug-in registry. */
      harness_plugins_data.count--;
      /* Reallocate plug-in registry. */
      if (0 == harness_plugins_data.count) {
        free(harness_plugins_data.plugins);
        harness_plugins_data.plugins = NULL;
      } else if (NULL == (plugins =
                          realloc(harness_plugins_data.plugins,
                                  harness_plugins_data.count *
                           sizeof(harness_plugins_data.plugins[0])))) {
        HARNESS_LOG((HARNESS_WARN(unable to reallocate plug-in registry)))
        errno = 0;
      } else {
        harness_plugins_data.plugins = plugins;
      }
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno3 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno3;
        harness_syserr();
      }
      /* Deallocate path. */
      free(path);
      /* Deallocate name duplicate. */
      free(namedup);
      errno = errno2;
      return -1;
    }
  }
  /* Randomly generate a unique non-zero 32-bit handle. */
  harness_plugins_data.plugins[index1].instances[index2].handle = 0;
  do {
    *handle = random() & 0xFFFFFFFF;
    for (index3 = 0;
         index3 < harness_plugins_data.count;
         index3++) {
      for (index4 = 0;
           index4 < harness_plugins_data.plugins[index3].count;
           index4++) {
        if (*handle ==
            harness_plugins_data.plugins[index3].instances[index4].handle) {
          break;
        }
      }
      if (index4 < harness_plugins_data.plugins[index3].count) {
        break;
      }
    }
  } while (index3 < harness_plugins_data.count);
  harness_plugins_data.plugins[index1].instances[index2].handle = *handle;
  /* Check if plug-in needs to be initialized. */
  if (NULL != init) {
    /* Reset ready flag. */
    harness_plugins_data.plugins[index1].instances[index2].ready = 0;
  } else {
    /* Set ready flag. */
    harness_plugins_data.plugins[index1].instances[index2].ready = 1;
  }
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check if plug-in does not need to be initialized. */
  if (NULL == init) {
    HARNESS_LOG((HARNESS_INFO(no init for %s plug-in), name))
    return 0;
  }
  /* Initialize plug-in. */
  if (NULL != init2) {
    if (0 != init2(*handle)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to initialize plug-in)))
      /* Unload plug-in. */
      if (0 != harness_plugins_unload(*handle)) {
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      }
      errno = errno2;
      return -1;
    }
  } else {
    if (0 != init(*handle)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to initialize plug-in)))
      /* Unload plug-in. */
      if (0 != harness_plugins_unload(*handle)) {
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      }
      errno = errno2;
      return -1;
    }
  }
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Search for plug-in in registry. */
  for (index1 = 0, index2 = 0;
       index1 < harness_plugins_data.count;
       index1++) {
    for (index2 = 0;
         index2 < harness_plugins_data.plugins[index1].count;
         index2++) {
      if (*handle ==
          harness_plugins_data.plugins[index1].instances[index2].handle) {
        break;
      }
    }
    if (index2 < harness_plugins_data.plugins[index1].count) {
      break;
    }
  }
  if (index1 == harness_plugins_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(unable to find plug-in in registry)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Set ready flag. */
  harness_plugins_data.plugins[index1].instances[index2].ready = 1;
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Loads a plug-in.
 *
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * first    = The first version supported of the plug-in to load.
 * age      = The number of followup versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_load3"
int harness_plugins_load3 (unsigned int *handle,
                        const char   *name,
                        unsigned int  first,
                        unsigned int  age,
                        unsigned int  revision,
                        unsigned int  flags) {
  int                  error;
  void                *module,  *instances, *plugins;
  char                *namedup, *prefix,    *path,   *func, *str;
  unsigned int         namelen,  prefixlen,  pathlen, funclen,
                       index1, index2, index3, index4;
  harness_plugins_init_t  init, init2;
  harness_plugins_fini_t  fini, fini2;
#ifdef DEBUG
  /* Check handle parameter. */
  if (NULL == handle) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(handle parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check name parameter. */
  if ('/' == name[0]) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name starts with '/')))
    errno = EINVAL;
    return -1;
  }
  if (NULL != strstr(name, "/..")) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter contains "/..")))
    errno = EINVAL;
    return -1;
  }
  /* Allocate name duplicate. */
  namelen = strlen(name);
  /* Check for .so ending. */
  if ((2 < namelen)&&(0 == strcmp(name + namelen - 3, ".so"))) {
    /* Remove .so ending. */
    namelen -= 3;
  }
  /* Check for .debug ending. */
  if ((5 < namelen)&&(0 == strcmp(name + namelen - 6, ".debug"))) {
    /* Remove .debug ending. */
    namelen -= 6;
  }
  if (NULL == (namedup = (char*)malloc((namelen + 1) * sizeof(char)))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to allocate name duplicate)))
    errno = errno2;
    return -1;
  }
  /* Compose name duplicate. */
  memcpy(namedup, name, namelen);
  namedup[namelen] = '\0';
  /* Allocate init/fini function name. */
  if (NULL == (func = (char*)malloc((namelen + 6) * sizeof(char)))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to allocate name duplicate)))
    errno = errno2;
    return -1;
  }
  /* Compose init/fini function name. */
  memcpy(func, namedup, namelen + 1);
  /* Remove lib prefix from function name. */
  if (NULL == (str = strrchr(func, '/'))) {
    str = func - 1;
  }
  str++;
  if (0 == strncmp(str, "lib", 3)) {
    memmove(str, str + 3, namelen - 2 - (str - func));
    funclen = namelen - 3;
  } else {
    funclen = namelen;
  }
  /* Replace unusual characters in function name. */
  for (index1 = 0; index1 < funclen; index1++) {
    if ((('0' > func[index1])||('9' < func[index1]))&&
        (('a' > func[index1])||('z' < func[index1]))&&
        (('A' > func[index1])||('Z' < func[index1]))) {
      func[index1] = '_';
    }
  }
  /* Allocate path. */
  prefix    =        HARNESS_STRING(HARNESS_LOADDIR);
  prefixlen = strlen(HARNESS_STRING(HARNESS_LOADDIR));
  pathlen   = prefixlen + namelen + 74;
  if (NULL == (path = (char*)malloc(pathlen * sizeof(char)))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to allocate path)))
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = errno2;
    return -1;
  }
  /* Compose path. */
#ifdef    DEBUG
  sprintf(path, "%s/%s.%u.%u.%u.debug.so",
          prefix, namedup, first, age, revision);
#else  /* DEBUG */
  sprintf(path, "%s/%s.%u.%u.%u.so",
          prefix, namedup, first, age, revision);
#endif /* DEBUG */
  /* Load plug-in. */
  if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_LOCAL))) {
    HARNESS_LOG((
      HARNESS_INFO(unable to load %s plug-in: %s), name, dlerror()))
    int errno2 = errno;
    /* Recompose path. */
#ifdef    DEBUG
    sprintf(path, "%s/%s.%u.%u.%u.so",
            prefix, namedup, first, age, revision);
#else  /* DEBUG */
    sprintf(path, "%s/%s.%u.%u.%u.debug.so",
            prefix, namedup, first, age, revision);
#endif /* DEBUG */
    /* Load plug-in. */
    if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_LOCAL))) {
      HARNESS_LOG((HARNESS_WARN(unable to load %s plug-in: %s), name, dlerror()))
      /* Deallocate path. */
      free(path);
      /* Deallocate init/fini function name. */
      free(func);
      /* Deallocate name duplicate. */
      free(namedup);
      errno = errno2;
      return -1;
    }
  }
  /* Resolve init function name. */
  memcpy(func + funclen, "_init", 6);
  init  = dlsym(module, func);
  init2 = NULL;
  /* Resolve fini function name. */
  memcpy(func + funclen, "_fini", 6);
  fini  = dlsym(module, func);
  fini2 = NULL;
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    /* Unload plug-in. */
    if (0 != dlclose(module)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = errno2;
      harness_syserr();
    }
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Deallocate name duplicate. */
    free(namedup);
    return -1;
  }
  /* Check for closing plug-in loader. */
  if (1 == harness_plugins_data.closing) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader is closing)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    /* Unload plug-in. */
    if (0 != dlclose(module)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = errno2;
      harness_syserr();
    }
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Deallocate name duplicate. */
    free(namedup);
    errno = EACCES;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    /* Unload plug-in. */
    if (0 != dlclose(module)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      errno = errno2;
      harness_syserr();
    }
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Deallocate name duplicate. */
    free(namedup);
    return -1;
  }
  /* Search for path in plug-in registry. */
  for (index1 = 0; index1 < harness_plugins_data.count; index1++) {
    if (0 == strcmp(harness_plugins_data.plugins[index1].path, path)) {
      break;
    }
  }
  /* Check if path is in plug-in registry. */
  if (index1 < harness_plugins_data.count) {
    /* Deallocate init/fini function name. */
    free(func);
    /* Deallocate path. */
    free(path);
    /* Compare names. */
    if (0 != strcmp(harness_plugins_data.plugins[index1].name, namedup)) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded under different name)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      /* Deallocate name duplicate. */
      free(namedup);
      errno = EACCES;
      return -1;
    }
    /* Deallocate name duplicate. */
    free(namedup);
    /* Compare modules. */
    if (harness_plugins_data.plugins[index1].module != module) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded under different module)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Compare init symbols. */
    if (harness_plugins_data.plugins[index1].init != init) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded with different init)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Compare fini symbols. */
    if (harness_plugins_data.plugins[index1].fini != fini) {
      errno = EACCES;
      HARNESS_LOG((HARNESS_WARN(plug-in loaded with different fini)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EACCES;
      return -1;
    }
    /* Check plug-in unique flag. */
    if ((1 == harness_plugins_data.plugins[index1].unique)||
        (HARNESS_PLUGINS_UNIQUE == (HARNESS_PLUGINS_UNIQUE & flags))) {
      errno = EALREADY;
      HARNESS_LOG((HARNESS_WARN(unable to load plug-in more than once)))
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno2;
        harness_syserr();
      }
      errno = EALREADY;
      return -1;
    }
    /* Increase instances array. */
    index2 = harness_plugins_data.plugins[index1].count;
    harness_plugins_data.plugins[index1].count++;
    /* Reallocate instances array. */
    if (NULL == (instances =
                 realloc(harness_plugins_data.plugins[index1].instances,
                         harness_plugins_data.plugins[index1].count *
                  sizeof(harness_plugins_data.plugins[0].instances[0])))) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to reallocate instances array)))
      harness_plugins_data.plugins[index1].count--;
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno3 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno3;
        harness_syserr();
      }
      errno = errno2;
      return -1;
    }
    harness_plugins_data.plugins[index1].instances = instances;
  } else {
    /* Check if name appears in plug-in registry. */
    index1 = 0;
    do {
      /* Search for name in plug-in registry. */
      for (; index1 < harness_plugins_data.count; index1++) {
        if (0 == strcmp(harness_plugins_data.plugins[index1].name, name)) {
          break;
        }
      }
      /* Check if name is in plug-in registry. */
      if (index1 < harness_plugins_data.count) {
        /* Check plug-in unique flag. */
        if ((1 == harness_plugins_data.plugins[index1].unique)||
            (HARNESS_PLUGINS_UNIQUE == (HARNESS_PLUGINS_UNIQUE & flags))) {
          errno = EALREADY;
          HARNESS_LOG((HARNESS_WARN(unable to load plug-in more than once)))
          /* Unlock plug-in loader mutex. */
          if (0 != (error =
                    pthread_mutex_unlock(&harness_plugins_data.mutex))) {
            errno = error;
            HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
            errno = error;
            harness_syserr();
          }
          /* Unload plug-in. */
          if (0 != dlclose(module)) {
            int errno2 = errno;
            HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
            errno = errno2;
            harness_syserr();
          }
          /* Deallocate init/fini function name. */
          free(func);
          /* Deallocate path. */
          free(path);
          /* Deallocate name duplicate. */
          free(namedup);
          errno = EALREADY;
          return -1;
        }
        index1++;
      }
    } while (index1 < harness_plugins_data.count);
    /* Check load plug-in globally flag to reload plug-in globally. */
    if ((HARNESS_PLUGINS_EXPORT == (HARNESS_PLUGINS_EXPORT & flags))) {
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno2 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        harness_syserr();
        /* Unlock plug-in loader mutex. */
        if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
          errno = error;
          HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
          errno = error;
          harness_syserr();
        }
        /* Deallocate path. */
        free(path);
        /* Deallocate init/fini function name. */
        free(func);
        /* Deallocate name duplicate. */
        free(namedup);
        errno = errno2;
        return -1;
      }
      /* Compose path. */
#ifdef    DEBUG
      sprintf(path, "%s/%s.debug.so", prefix, namedup);
#else  /* DEBUG */
      sprintf(path, "%s/%s.so", prefix, namedup);
#endif /* DEBUG */
      /* Load plug-in. */
      if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_GLOBAL))) {
        HARNESS_LOG((HARNESS_INFO(unable to reload %s plug-in globally: %s),\
                 name, dlerror()))
        int errno2 = errno;
        /* Recompose path. */
#ifdef    DEBUG
        sprintf(path, "%s/%s.so", prefix, namedup);
#else  /* DEBUG */
        sprintf(path, "%s/%s.debug.so", prefix, namedup);
#endif /* DEBUG */
        /* Load plug-in. */
        if (NULL == (module = dlopen(path, RTLD_LAZY|RTLD_GLOBAL))) {
          HARNESS_LOG((
            HARNESS_WARN(unable to reload %s plug-in globally: %s),\
                   name, dlerror()))
          /* Unlock plug-in loader mutex. */
          if (0 != (error =
                    pthread_mutex_unlock(&harness_plugins_data.mutex))) {
            HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
            harness_syserr();
          }
          /* Deallocate path. */
          free(path);
          /* Deallocate init/fini function name. */
          free(func);
          /* Deallocate name duplicate. */
          free(namedup);
          errno = errno2;
          return -1;
        }
      }
      /* Resolve init function name. */
      memcpy(func + namelen, "_init", 6);
      init2 = dlsym(module, func);
      /* Resolve fini function name. */
      memcpy(func + namelen, "_fini", 6);
      fini2 = dlsym(module, func);
      /* Deallocate init/fini function name. */
      free(func);
    } else {
      /* Deallocate init/fini function name. */
      free(func);
    }
    /* Increase plug-in registry. */
    harness_plugins_data.count++;
    /* Reallocate plug-in registry. */
    if (NULL == (plugins = realloc(harness_plugins_data.plugins,
                                   harness_plugins_data.count *
                            sizeof(harness_plugins_data.plugins[0])))) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to reallocate plug-in registry)))
      harness_plugins_data.count--;
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno3 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno3;
        harness_syserr();
      }
      /* Deallocate path. */
      free(path);
      /* Deallocate name duplicate. */
      free(namedup);
      errno = errno2;
      return -1;
    }
    harness_plugins_data.plugins = plugins;
    /* Save plug-in registry entry. */
    harness_plugins_data.plugins[index1].path   = path;
    harness_plugins_data.plugins[index1].name   = namedup;
    harness_plugins_data.plugins[index1].module = module;
    harness_plugins_data.plugins[index1].init   = init;
    harness_plugins_data.plugins[index1].fini   = fini;
    harness_plugins_data.plugins[index1].init2  = init2;
    harness_plugins_data.plugins[index1].fini2  = fini2;
    harness_plugins_data.plugins[index1].unique =
      (HARNESS_PLUGINS_UNIQUE == (HARNESS_PLUGINS_UNIQUE & flags))?1:0;
    /* Increase instances array. */
    index2 = 0;
    harness_plugins_data.plugins[index1].count = 1;
    /* Allocate instances array. */
    if (NULL == (harness_plugins_data.plugins[index1].instances = malloc(
          sizeof(harness_plugins_data.plugins[index1].instances[0])))) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to allocate instances array)))
      /* Decrease plug-in registry. */
      harness_plugins_data.count--;
      /* Reallocate plug-in registry. */
      if (0 == harness_plugins_data.count) {
        free(harness_plugins_data.plugins);
        harness_plugins_data.plugins = NULL;
      } else if (NULL == (plugins =
                          realloc(harness_plugins_data.plugins,
                                  harness_plugins_data.count *
                           sizeof(harness_plugins_data.plugins[0])))) {
        HARNESS_LOG((HARNESS_WARN(unable to reallocate plug-in registry)))
      } else {
        harness_plugins_data.plugins = plugins;
      }
      /* Unlock plug-in loader mutex. */
      if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
        errno = error;
        harness_syserr();
      }
      /* Unload plug-in. */
      if (0 != dlclose(module)) {
        int errno3 = errno;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = errno3;
        harness_syserr();
      }
      /* Deallocate path. */
      free(path);
      /* Deallocate name duplicate. */
      free(namedup);
      errno = errno2;
      return -1;
    }
  }
  /* Randomly generate a unique non-zero 32-bit handle. */
  harness_plugins_data.plugins[index1].instances[index2].handle = 0;
  do {
    *handle = random() & 0xFFFFFFFF;
    for (index3 = 0;
         index3 < harness_plugins_data.count;
         index3++) {
      for (index4 = 0;
           index4 < harness_plugins_data.plugins[index3].count;
           index4++) {
        if (*handle ==
            harness_plugins_data.plugins[index3].instances[index4].handle) {
          break;
        }
      }
      if (index4 < harness_plugins_data.plugins[index3].count) {
        break;
      }
    }
  } while (index3 < harness_plugins_data.count);
  harness_plugins_data.plugins[index1].instances[index2].handle = *handle;
  /* Check if plug-in needs to be initialized. */
  if (NULL != init) {
    /* Reset ready flag. */
    harness_plugins_data.plugins[index1].instances[index2].ready = 0;
  } else {
    /* Set ready flag. */
    harness_plugins_data.plugins[index1].instances[index2].ready = 1;
  }
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check if plug-in does not need to be initialized. */
  if (NULL == init) {
    HARNESS_LOG((HARNESS_INFO(no init for %s plug-in), name))
    return 0;
  }
  /* Initialize plug-in. */
  if (NULL != init2) {
    if (0 != init2(*handle)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to initialize plug-in)))
      /* Unload plug-in. */
      if (0 != harness_plugins_unload(*handle)) {
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      }
      errno = errno2;
      return -1;
    }
  } else {
    if (0 != init(*handle)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to initialize plug-in)))
      /* Unload plug-in. */
      if (0 != harness_plugins_unload(*handle)) {
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
      }
      errno = errno2;
      return -1;
    }
  }
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Search for plug-in in registry. */
  for (index1 = 0, index2 = 0;
       index1 < harness_plugins_data.count;
       index1++) {
    for (index2 = 0;
         index2 < harness_plugins_data.plugins[index1].count;
         index2++) {
      if (*handle ==
          harness_plugins_data.plugins[index1].instances[index2].handle) {
        break;
      }
    }
    if (index2 < harness_plugins_data.plugins[index1].count) {
      break;
    }
  }
  if (index1 == harness_plugins_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(unable to find plug-in in registry)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Set ready flag. */
  harness_plugins_data.plugins[index1].instances[index2].ready = 1;
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Loads a plug-in.
 *
 * handle   = The plug-in handle (return). Plug-in handles are unique.
 * name     = The plug-in name. Plug-in names are unique and do
 *            not contain '/..' and do not start with '/'.
 * current  = The current version supported of the plug-in to load.
 * age      = The number of previous versions supported of the plug-in to load.
 * revision = The revision of the plug-in to load.
 * flags    = The plug-in loading flags, e.g. HARNESS_PLUGINS_UNIQUE.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_load4"
int harness_plugins_load4 (unsigned int *handle,
                        const char   *name,
                        unsigned int  current,
                        unsigned int  age,
                        unsigned int  revision,
                        unsigned int  flags) {
#ifdef DEBUG
  /* Check current parameter. */
  if (current < age) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(current is less than age)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  if (0 != harness_plugins_load3(handle, name, current - age,
                              age, revision, flags)) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to load plug-in)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Gets the module of a plug-in (for dlsym).
 *
 * module = The plug-in module (return).
 * handle = The plug-in handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_module"
int harness_plugins_module (void         **module,
                         unsigned int   handle) {
  int          error;
  unsigned int index1, index2;
#ifdef DEBUG
  /* Check module parameter. */
  if (NULL == module) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(module parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check for closing plug-in loader. */
  if (1 == harness_plugins_data.closing) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader is closing)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  for (index1 = 0, index2 = 0;
       index1 < harness_plugins_data.count;
       index1++) {
    for (index2 = 0;
         index2 < harness_plugins_data.plugins[index1].count;
         index2++) {
      if (handle ==
          harness_plugins_data.plugins[index1].instances[index2].handle) {
        break;
      }
    }
    if (index2 < harness_plugins_data.plugins[index1].count) {
      break;
    }
  }
  if (index1 == harness_plugins_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(unable to find plug-in in registry)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Save plug-in module. */
  *module = harness_plugins_data.plugins[index1].module;
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Gets the symbol of a plug-in corresponding to a name (using dlsym).
 *
 * symbol = The plug-in symbol that corresponds to the name(return).
 * handle = The plug-in handle.
 * name   = The plug-in symbol name to resolve.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_symbol"
int harness_plugins_symbol (void         **symbol,
                         unsigned int   handle,
                         const char    *name) {
  int          error;
  unsigned int index1, index2;
#ifdef DEBUG
  /* Check symbol parameter. */
  if (NULL == symbol) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(symbol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check for closing plug-in loader. */
  if (1 == harness_plugins_data.closing) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader is closing)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  for (index1 = 0, index2 = 0;
       index1 < harness_plugins_data.count;
       index1++) {
    for (index2 = 0;
         index2 < harness_plugins_data.plugins[index1].count;
         index2++) {
      if (handle ==
          harness_plugins_data.plugins[index1].instances[index2].handle) {
        break;
      }
    }
    if (index2 < harness_plugins_data.plugins[index1].count) {
      break;
    }
  }
  if (index1 == harness_plugins_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(unable to find plug-in in registry)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Resolve the symbol. */
  *symbol = dlsym(harness_plugins_data.plugins[index1].module, name);
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Gets the symbol of a plug-in corresponding to a name (using dlsym).
 * Prepends <plug-in name>_ (without any lib prefix) to the symbol name.
 *
 * symbol = The plug-in symbol that corresponds to the name(return).
 * handle = The plug-in handle.
 * name   = The plug-in symbol name to resolve.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_symbol2"
int harness_plugins_symbol2 (void         **symbol,
                          unsigned int   handle,
                          const char    *name) {
  int           error;
  char         *head, *symbolname;
  unsigned int  index1, index2, headlen, taillen;
#ifdef DEBUG
  /* Check symbol parameter. */
  if (NULL == symbol) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(symbol parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check for closing plug-in loader. */
  if (1 == harness_plugins_data.closing) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader is closing)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  for (index1 = 0, index2 = 0;
       index1 < harness_plugins_data.count;
       index1++) {
    for (index2 = 0;
         index2 < harness_plugins_data.plugins[index1].count;
         index2++) {
      if (handle ==
          harness_plugins_data.plugins[index1].instances[index2].handle) {
        break;
      }
    }
    if (index2 < harness_plugins_data.plugins[index1].count) {
      break;
    }
  }
  if (index1 == harness_plugins_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(unable to find plug-in in registry)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  head    = harness_plugins_data.plugins[index1].name;
  headlen = strlen(head);
  taillen = strlen(name);
  /* Allocate symbol name. */
  if (NULL == (symbolname = (char*)malloc((headlen + taillen + 2) *
                                          sizeof(char)))) {
    int errno2 = errno;
    HARNESS_LOG((HARNESS_WARN(unable to allocate symbol name)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Copy plug-in name into symbol name as prefix. */
  memcpy(symbolname, head, headlen);
  symbolname[headlen] = '\0';
  /* Remove lib prefix from symbol name. */
  if (NULL == (head = strrchr(symbolname, '/'))) {
    head = symbolname - 1;
  }
  head++;
  if (0 == strncmp(head, "lib", 3)) {
    memmove(head, head + 3, headlen - 2 - (head - symbolname));
    headlen -= 3;
  }
  /* Append symbol name to plug-in name. */
  symbolname[headlen] = '_';
  memcpy(symbolname + headlen + 1, name, taillen);
  headlen += taillen + 1;
  symbolname[headlen] = '\0';
  /* Replace unusual characters in symbol name. */
  for (index2 = 0; index2 < headlen; index2++) {
    if ((('0' > symbolname[index2])||('9' < symbolname[index2]))&&
        (('a' > symbolname[index2])||('z' < symbolname[index2]))&&
        (('A' > symbolname[index2])||('Z' < symbolname[index2]))) {
      symbolname[index2] = '_';
    }
  }
  /* Resolve the symbol. */
  *symbol = dlsym(harness_plugins_data.plugins[index1].module, symbolname);
  /* Deallocate symbol name. */
  free(symbolname);
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Gets the name of a plug-in.
 *
 * name   = The plug-in name (return).
 * handle = The plug-in handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_name"
int harness_plugins_name (const char   **name,
                       unsigned int   handle) {
  int          error;
  unsigned int index1, index2;
#ifdef DEBUG
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    HARNESS_LOG((HARNESS_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check for closing plug-in loader. */
  if (1 == harness_plugins_data.closing) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader is closing)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  for (index1 = 0, index2 = 0;
       index1 < harness_plugins_data.count;
       index1++) {
    for (index2 = 0;
         index2 < harness_plugins_data.plugins[index1].count;
         index2++) {
      if (handle ==
          harness_plugins_data.plugins[index1].instances[index2].handle) {
        break;
      }
    }
    if (index2 < harness_plugins_data.plugins[index1].count) {
      break;
    }
  }
  if (index1 == harness_plugins_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(unable to find plug-in in registry)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Return the plug-in name. */
  *name = harness_plugins_data.plugins[index1].name;
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Unloads a plug-in.
 *
 * handle = The plug-in handle.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "harness_plugins_unload"
int harness_plugins_unload (unsigned int handle) {
  int                  error;
  void                *module, *instances, *plugins;
  unsigned int         index1, index2;
  harness_plugins_fini_t  fini, fini2;
  /* Lock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_lock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to lock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized plug-in loader. */
  if (0 == harness_plugins_data.ready) {
    errno = EACCES;
    HARNESS_LOG((HARNESS_WARN(plug-in loader finalized)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Search for plug-in in registry. */
  for (index1 = 0, index2 = 0;
       index1 < harness_plugins_data.count;
       index1++) {
    for (index2 = 0;
         index2 < harness_plugins_data.plugins[index1].count;
         index2++) {
      if (handle ==
          harness_plugins_data.plugins[index1].instances[index2].handle) {
        break;
      }
    }
    if (index2 < harness_plugins_data.plugins[index1].count) {
      break;
    }
  }
  if (index1 == harness_plugins_data.count) {
    errno = ESRCH;
    HARNESS_LOG((HARNESS_WARN(unable to find plug-in in registry)))
    /* Unlock plug-in loader mutex. */
    if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
      errno = error;
      HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
      errno = error;
      harness_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Check ready flag. */
  if (1 == harness_plugins_data.plugins[index1].instances[index2].ready) {
    /* Save fini symbol and module. */
    fini   = harness_plugins_data.plugins[index1].fini;
    fini2  = harness_plugins_data.plugins[index1].fini2;
    module = harness_plugins_data.plugins[index1].module;
  } else {
    /* Save module. */
    fini   = NULL;
    fini2  = NULL;
    module = harness_plugins_data.plugins[index1].module;
  }
  /* Remove instance from instances array. */
  harness_plugins_data.plugins[index1].count--;
  memmove(harness_plugins_data.plugins[index1].instances + index2,
          harness_plugins_data.plugins[index1].instances + index2 + 1,
         (harness_plugins_data.plugins[index1].count     - index2) *
   sizeof(harness_plugins_data.plugins[0].instances[0]));
  /* Reallocate instances array. */
  if (0 == harness_plugins_data.plugins[index1].count) {
    free(harness_plugins_data.plugins[index1].instances);
    harness_plugins_data.plugins[index1].instances = NULL;
  } else  if (NULL ==
              (instances =
               realloc(harness_plugins_data.plugins[index1].instances,
                       harness_plugins_data.plugins[index1].count *
                sizeof(harness_plugins_data.plugins[0].instances[0])))) {
    HARNESS_LOG((HARNESS_WARN(unable to reallocate instances array)))
    errno = 0;
  } else {
    harness_plugins_data.plugins[index1].instances = instances;
  }
  /* Check if plug-in needs to be removed from plug-in registry. */
  if (0 == harness_plugins_data.plugins[index1].count) {
    /* Deallocate path. */
    free(harness_plugins_data.plugins[index1].path);
    /* Deallocate name. */
    free(harness_plugins_data.plugins[index1].name);
    /* Remove plug-in from plug-in registry. */
    harness_plugins_data.count--;
    memmove(harness_plugins_data.plugins + index1,
            harness_plugins_data.plugins + index1 + 1,
           (harness_plugins_data.count   - index1) *
            sizeof(harness_plugins_data.plugins[0]));
    /* Reallocate plug-in registry. */
    if (0 == harness_plugins_data.count) {
      free(harness_plugins_data.plugins);
      harness_plugins_data.plugins = NULL;
    } else if (NULL == (plugins = realloc(harness_plugins_data.plugins,
                                          harness_plugins_data.count *
                                   sizeof(harness_plugins_data.plugins[0])))) {
      HARNESS_LOG((HARNESS_WARN(unable to reallocate plug-in registry)))
      errno = 0;
    } else {
      harness_plugins_data.plugins = plugins;
    }
  }
  /* Unlock plug-in loader mutex. */
  if (0 != (error = pthread_mutex_unlock(&harness_plugins_data.mutex))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unlock plug-in loader mutex)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  /* Finalize plug-in. */
  if (NULL != fini2) {
    if (0 != fini2(handle)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in)))
      /* Unload plug-in. */
      if (0 != (error = dlclose(module))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = error;
        harness_syserr();
      }
      errno = errno2;
      return -1;
    }
  } else if (NULL != fini) {
    if (0 != fini(handle)) {
      int errno2 = errno;
      HARNESS_LOG((HARNESS_WARN(unable to finalize plug-in)))
      /* Unload plug-in. */
      if (0 != (error = dlclose(module))) {
        errno = error;
        HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
        errno = error;
        harness_syserr();
      }
      errno = errno2;
      return -1;
    }
  }
  /* Unload plug-in. */
  if (0 != (error = dlclose(module))) {
    errno = error;
    HARNESS_LOG((HARNESS_WARN(unable to unload plug-in)))
    errno = error;
    harness_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
