/******************************************************************************
 *
 * Source file for rmix method (object method descriptor) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Creates an object method descriptor.
 *
 * method = The object method descriptor to create (return).
 * name   = The object method name (cannot be NULL).
 * output = The object method output signature (cannot be NULL).
 * input  = The object method input  signature (cannot be NULL).
 * call   = The object method function (cannot be NULL).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_method_create"
int rmix_method_create (rmix_method_t      **method,
                        const char          *name,
                        const char          *output,
                        const char          *input,
                        rmix_method_call_t   call) {
#ifdef DEBUG
  /* Check method parameter. */
  if (NULL == method) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check name parameter. */
  if (NULL == name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check output parameter. */
  if (NULL == output) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(output parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check input parameter. */
  if (NULL == input) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(input parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check call parameter. */
  if (NULL == call) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(call parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Allocate object method descriptor. */
  if (NULL == (*method = malloc(sizeof(rmix_method_t)))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate object method descriptor)))
    errno = errno2;
    return -1;
  }
  /* Duplicate object method name. */
  if (NULL == ((*method)->name = strdup(name))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method name)))
    /* Deallocate object method descriptor. */
    free(*method);
    *method = NULL;
    errno = errno2;
    return -1;
  }
  /* Duplicate object method output signature. */
  if (NULL == ((*method)->output = strdup(output))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method output signature)))
    /* Deallocate object method name. */
    free((*method)->name);
    /* Deallocate object method descriptor. */
    free(*method);
    *method = NULL;
    errno = errno2;
    return -1;
  }
  /* Duplicate object method input signature. */
  if (NULL == ((*method)->input = strdup(input))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object method input signature)))
    /* Deallocate object method output signature. */
    free((*method)->output);
    /* Deallocate object method name. */
    free((*method)->name);
    /* Deallocate object method descriptor. */
    free(*method);
    *method = NULL;
    errno = errno2;
    return -1;
  }
  /* Copy object method function. */
  (*method)->call = call;
  return 0;
}

/*
 * Destroys an object method descriptor.
 *
 * method = The object method descriptor to destroy (also return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_method_destroy"
int rmix_method_destroy (rmix_method_t **method) {
#ifdef DEBUG
  /* Check method parameter. */
  if (NULL == method) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check *method parameter. */
  if (NULL == *method) {
    return 0;
  }
  /* Deallocate object method input signature. */
  free((*method)->input);
  /* Deallocate object method output signature. */
  free((*method)->output);
  /* Deallocate object method name. */
  free((*method)->name);
  /* Deallocate object method descriptor. */
  free(*method);
  *method = NULL;
  return 0;
}

/*
 * Duplicates an object method descriptor.
 *
 * method1 = The object method descriptor to create (return).
 * method2 = The object method descriptor to duplicate.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_method_duplicate"
int rmix_method_duplicate (rmix_method_t       **method1,
                           const rmix_method_t  *method2) {
#ifdef DEBUG
  /* Check method1 parameter. */
  if (NULL == method1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method2 parameter. */
  if (NULL == method2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Create object method descriptor. */
  if (0 != rmix_method_create(method1, method2->name, method2->output,
                              method2->input, method2->call)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to create object method descriptor)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Compares two object method descriptors.
 *
 * diff    = The difference between both object method descriptors (return).
 *           diff > 0 if method1 < method2,
 *           diff < 0 if method1 > method2, and
 *           diff = 0 if method1 = method2
 * method1 = The object method descriptor to compare against.
 * method2 = The object method descriptor to compare with.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_method_compare"
int rmix_method_compare (int                    *diff,
                         const rmix_method_t *method1,
                         const rmix_method_t *method2) {
#ifdef DEBUG
  /* Check diff parameter. */
  if (NULL == diff) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(diff parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method1 parameter. */
  if (NULL == method1) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method1 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method1->name parameter. */
  if (NULL == method1->name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method1->name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method1->output parameter. */
  if (NULL == method1->output) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method1->output parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method1->input parameter. */
  if (NULL == method1->input) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method1->input parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method1->call parameter. */
  if (NULL == method1->call) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method1->call parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method2 parameter. */
  if (NULL == method2) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method2 parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method2->name parameter. */
  if (NULL == method2->name) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method2->name parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method2->output parameter. */
  if (NULL == method2->output) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method2->output parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method2->input parameter. */
  if (NULL == method2->input) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method2->input parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check method2->call parameter. */
  if (NULL == method2->call) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(method2->call parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Compare object method descriptors. */
  if ((0 != (*diff = strcmp(method1->name  , method2->name  )))&&
      (0 != (*diff = strcmp(method1->output, method2->output)))&&
      (0 != (*diff = strcmp(method1->input , method2->input )))) {
    *diff = method2->call - method1->call;
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
