/******************************************************************************
 *
 * Source file for rmix objects (object registry) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Object registry data type. */
typedef struct {
  pthread_mutex_t     mutex;      /* mutex                    */
  unsigned int        ready;      /* initialized flag         */
  unsigned int        count;      /* object registry count    */
  struct {
    void             *object;     /* object data              */
    unsigned int      objectid;   /* object id                */
    unsigned int      references; /* object reference counter */
    rmix_interface_t *interface;  /* object interface         */
  }                  *objects;    /* object registry array    */
} rmix_objects_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the object registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_objects_init ();

/*
 * Finalizes the object registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_objects_fini ();

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Object registry data. */
rmix_objects_data_t rmix_objects_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex                 */
  0,                         /* initialized flag      */
  0,                         /* object registry count */
  NULL                       /* object registry array */
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the object registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_objects_init"
int rmix_objects_init () {
  int error;
  RMIX_LOG((RMIX_INFO(object registry startup)))
  /* Lock object registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for initialized object registry. */
  if (1 == rmix_objects_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(object registry already initialized)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Initialize object registry. */
  rmix_objects_data.count   = 0;
  rmix_objects_data.objects = NULL;
  /* Mark object registry initialized. */
  rmix_objects_data.ready   = 1;
  /* Unlock object registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
    errno = error;
    rmix_syserr();
    /* Mark object registry finalized. */
    rmix_objects_data.ready = 0;
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the object registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_objects_fini"
int rmix_objects_fini () {
  int          error, result, errno2;
  unsigned int index;
  RMIX_LOG((RMIX_INFO(object registry shutdown)))
  result = 0;
  errno2 = 0;
  /* Lock object registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized object registry. */
  if (0 == rmix_objects_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(object registry already finalized)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Finalize object registry. */
  for (index = 0; index < rmix_objects_data.count; index++) {
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(
               &rmix_objects_data.objects[index].interface)) {
      errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
      result = -1;
    }
  }
  if (NULL != rmix_objects_data.objects) {
    free(rmix_objects_data.objects);
    rmix_objects_data.objects = NULL;
  }
  rmix_objects_data.count = 0;
  /* Mark object registry finalized. */
  rmix_objects_data.ready = 0;
  /* Unlock object registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Adds an object to the object registry. The object may already exist in the
 * registry. In this case its reference counter is increased and the same
 * object id is returned. An object may register with a different interface. In
 * this case a new object id is returned.
 *
 * objectid  = The object id (return).
 * object    = The object to add (may be NULL).
 * interface = The object interface to add.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_objects_add"
int rmix_objects_add (unsigned int           *objectid,
                      void                   *object,
                      const rmix_interface_t *interface) {
  int                  error, diff;
  void                *objects;
  unsigned int         index1, index2;
  rmix_interface_t *interface2;
#ifdef DEBUG
  /* Check objectid parameter. */
  if (NULL == objectid) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(objectid parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Duplicate object interface. */
  if (0 != rmix_interface_duplicate(&interface2, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object interface)))
    errno = errno2;
    return -1;
  }
  /* Lock object registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock object registry mutex)))
    errno = error;
    rmix_syserr();
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(&interface2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
    }
    errno = error;
    return -1;
  }
  /* Check for finalized object registry. */
  if (0 == rmix_objects_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(object registry finalized)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(&interface2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
    }
    errno = EACCES;
    return -1;
  }
  /* Search for object in registry. */
  for (index1 = 0; index1 < rmix_objects_data.count; index1++) {
    /* Compare objects. */
    if (object == rmix_objects_data.objects[index1].object) {
      /* Compare interfaces. */
      if (0 != rmix_interface_compare(&diff, interface2,
                 rmix_objects_data.objects[index1].interface)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to compare interfaces)))
        /* Unlock object registry mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
          rmix_syserr();
        }
        /* Destroy object interface. */
        if (0 != rmix_interface_destroy(&interface2)) {
          RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
        }
        errno = errno2;
        return -1;
      }
      if (0 == diff) {
        break;
      }
    }
  }
  /* Check for object in registry. */
  if (index1 != rmix_objects_data.count) {
    /* Increase reference counter. */
    rmix_objects_data.objects[index1].references++;
    /* Save object id. */
    *objectid = rmix_objects_data.objects[index1].objectid;
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
      /* Destroy object interface. */
      if (0 != rmix_interface_destroy(&interface2)) {
        RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
      }
      errno = error;
      return -1;
    }
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(&interface2)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
      errno2 = errno;
      return -1;
    }
    return 0;
  }
  /* Increase object registry. */
  index1 = rmix_objects_data.count;
  rmix_objects_data.count++;
  /* Reallocate object registry. */
  if (NULL == (objects = realloc(rmix_objects_data.objects,
                                 rmix_objects_data.count *
                          sizeof(rmix_objects_data.objects[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate object registry)))
    /* Decrease object registry. */
    rmix_objects_data.count--;
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(&interface2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
    }
    errno = errno2;
    return -1;
  }
  rmix_objects_data.objects = objects;
  /* Save object registry entry. */
  rmix_objects_data.objects[index1].object     = object;
  rmix_objects_data.objects[index1].objectid     = 0;
  rmix_objects_data.objects[index1].references = 1;
  rmix_objects_data.objects[index1].interface  = interface2;
  /* Randomly generate a unique non-zero 16-bit object id. */
  do {
    *objectid = random() & 0xFFFF;
    for (index2 = 0; index2 < rmix_objects_data.count; index2++) {
      if (*objectid == rmix_objects_data.objects[index2].objectid) {
        break;
      }
    }
  } while (index2 < rmix_objects_data.count);
  rmix_objects_data.objects[index1].objectid = *objectid;
  /* Unlock object registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Adds an object to the object registry with a forced object id (if not already
 * used by another object). The object may already exist under this object id in
 * the registry. In this case its reference counter is increased. The
 * object interfaces must be equal for successful registration.
 *
 * objectid  = The object id.
 * object    = The object to add.
 * interface = The object interface to add.
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_objects_add2"
int rmix_objects_add2 (unsigned int            objectid,
                       void                   *object,
                       const rmix_interface_t *interface) {
  int                  error, diff;
  void                *objects;
  unsigned int         index1;
  rmix_interface_t *interface2;
#ifdef DEBUG
  /* Check interface parameter. */
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Check objectid parameter. */
  if (0xFFFF < objectid) {
    errno = E2BIG;
    RMIX_LOG((RMIX_WARN(objectid parameter is too big)))
    errno = E2BIG;
    return -1;
  }
  /* Duplicate object interface. */
  if (0 != rmix_interface_duplicate(&interface2, interface)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to duplicate object interface)))
    errno = errno2;
    return -1;
  }
  /* Lock object registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock object registry mutex)))
    errno = error;
    rmix_syserr();
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(&interface2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
    }
    errno = error;
    return -1;
  }
  /* Check for finalized object registry. */
  if (0 == rmix_objects_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(object registry finalized)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(&interface2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
    }
    errno = EACCES;
    return -1;
  }
  /* Search for object id in registry. */
  for (index1 = 0; index1 < rmix_objects_data.count; index1++) {
    if (index1 == rmix_objects_data.objects[index1].objectid) {
      break;
    }
  }
  /* Check for object id in registry. */
  if (index1 != rmix_objects_data.count) {
    /* Compare objects. */
    if (object == rmix_objects_data.objects[index1].object) {
      /* Compare interfaces. */
      if (0 != rmix_interface_compare(&diff, interface2,
                 rmix_objects_data.objects[index1].interface)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to compare interfaces)))
        /* Unlock object registry mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
          errno = error;
          rmix_syserr();
        }
        /* Destroy object interface. */
        if (0 != rmix_interface_destroy(&interface2)) {
          RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
        }
        errno = errno2;
        return -1;
      }
      if (0 == diff) {
        /* Increase reference counter. */
        rmix_objects_data.objects[index1].references++;
        /* Unlock object registry mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
          errno = error;
          rmix_syserr();
          /* Destroy object interface. */
          if (0 != rmix_interface_destroy(&interface2)) {
            RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
          }
          errno = error;
          return -1;
        }
        /* Destroy object interface. */
        if (0 != rmix_interface_destroy(&interface2)) {
          int errno2 = errno;
          RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
          errno = errno2;
          return -1;
        }
        return 0;
      }
      errno = EALREADY;
      RMIX_LOG((
        RMIX_WARN(\
          object id/object already registered with different interface)))
      /* Unlock object registry mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
        errno = error;
        rmix_syserr();
      }
      /* Destroy object interface. */
      if (0 != rmix_interface_destroy(&interface2)) {
        RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
      }
      errno = EALREADY;
      return -1;
    }
    errno = EALREADY;
    RMIX_LOG((RMIX_WARN(object id already registered with different object)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(&interface2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
    }
    errno = EALREADY;
    return -1;
  }
  /* Search for object in registry. */
  for (index1 = 0; index1 < rmix_objects_data.count; index1++) {
    /* Compare objects. */
    if (object == rmix_objects_data.objects[index1].object) {
      /* Compare interfaces. */
      if (0 != rmix_interface_compare(&diff, interface2,
                 rmix_objects_data.objects[index1].interface)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to compare interfaces)))
        /* Unlock object registry mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
          errno = error;
          rmix_syserr();
        }
        /* Destroy object interface. */
        if (0 != rmix_interface_destroy(&interface2)) {
          RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
        }
        errno = errno2;
        return -1;
      }
      if (0 == diff) {
        break;
      }
    }
  }
  /* Check for object in registry. */
  if (index1 != rmix_objects_data.count) {
    errno = EALREADY;
    RMIX_LOG((RMIX_WARN(object already registered with different object id)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(&interface2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
    }
    errno = EALREADY;
    return -1;
  }
  /* Increase object registry. */
  index1 = rmix_objects_data.count;
  rmix_objects_data.count++;
  /* Reallocate object registry. */
  if (NULL == (objects = realloc(rmix_objects_data.objects,
                                 rmix_objects_data.count *
                          sizeof(rmix_objects_data.objects[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to reallocate object registry)))
    /* Decrease object registry. */
    rmix_objects_data.count--;
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(&interface2)) {
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
    }
    errno = errno2;
    return -1;
  }
  rmix_objects_data.objects = objects;
  /* Save object registry entry. */
  rmix_objects_data.objects[index1].object     = object;
  rmix_objects_data.objects[index1].objectid   = objectid;
  rmix_objects_data.objects[index1].references = 1;
  rmix_objects_data.objects[index1].interface  = interface2;
  /* Unlock object registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Removes an object from the object registry. If its reference counter is not
 * zero, the object will remain in the object registry.
 *
 * objectid = The object id to remove.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_objects_remove"
int rmix_objects_remove (unsigned int objectid) {
  int           error, result, errno2;
  void         *objects;
  unsigned int  index;
  result = 0;
  errno2 = 0;
  /* Lock object registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized object registry. */
  if (0 == rmix_objects_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(object registry finalized)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Search for object id in registry. */
  for (index = 0; index < rmix_objects_data.count; index++) {
    if (objectid == rmix_objects_data.objects[index].objectid) {
      break;
    }
  }
  if (index == rmix_objects_data.count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(unable to find object id in registry)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Check references. */
  if (1 < rmix_objects_data.objects[index].references) {
    /* Reduce the number of references. */
    rmix_objects_data.objects[index].references--;
  } else {
    /* Destroy object interface. */
    if (0 != rmix_interface_destroy(
               &rmix_objects_data.objects[index].interface)) {
      errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to destroy object interface)))
      result = -1;
    }
    /* Decrease object registry. */
    rmix_objects_data.count--;
    memmove(rmix_objects_data.objects + index,
            rmix_objects_data.objects + index + 1,
           (rmix_objects_data.count   - index) *
            sizeof(rmix_objects_data.objects[0]));
    /* Reallocate object registry. */
    if (0 == rmix_objects_data.count) {
        free(rmix_objects_data.objects);
        rmix_objects_data.objects = NULL;
    } else if (NULL == (objects = realloc(rmix_objects_data.objects,
                                   sizeof(rmix_objects_data.objects[0]) *
                                          rmix_objects_data.count))) {
      RMIX_LOG((
        RMIX_INFO(unable to reallocate object registry)))
    } else {
      rmix_objects_data.objects = objects;
    }
  }
  /* Unlock object registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
    errno = error;
    rmix_syserr();
    return -1;
  }
  if (0 != errno2) {
    errno = errno2;
  }
  return result;
}

/*
 * Returns the object and interface of a registered object.
 *
 * objectid    = The object id.
 * object    = The object (return).
 * interface = The object interface (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_objects_get"
int rmix_objects_get (unsigned int             objectid,
                      void                   **object,
                      const rmix_interface_t **interface) {
  int          error;
  unsigned int index;
#ifdef DEBUG
  /* Check object parameter. */
  if (NULL == object) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(object parameter is null)))
    errno = EINVAL;
    return -1;
  }
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock object registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized object registry. */
  if (0 == rmix_objects_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(object registry finalized)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Search for object id in registry. */
  for (index = 0; index < rmix_objects_data.count; index++) {
    if (objectid == rmix_objects_data.objects[index].objectid) {
      break;
    }
  }
  if (index == rmix_objects_data.count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(unable to find object id in registry)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Copy object and interface. */
  *object    = rmix_objects_data.objects[index].object;
  *interface = rmix_objects_data.objects[index].interface;
  /* Unlock object registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Returns the object and interface of a registered object.
 *
 * remoteref = The object remote reference.
 * object    = The object (return).
 * interface = The object interface (return).
 * return    = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_objects_get2"
int rmix_objects_get2 (const rmix_remoteref_t  *remoteref,
                       void                   **object,
                       const rmix_interface_t **interface) {
  int          error;
  unsigned int objectid, index;
#ifdef DEBUG
  /* Check remoteref parameter. */
  if (NULL == remoteref) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(remoteref parameter is null)))
    errno = EINVAL;
    return -1;
  }
  /* Check object parameter. */
  if (NULL == object) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(object parameter is null)))
    errno = EINVAL;
    return -1;
  }
  if (NULL == interface) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(interface parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Get object hadle from remote reference. */
  if (0 != (error = rmix_remoteref_objectid(remoteref, &objectid))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get object hadle from remote reference)))
    errno = errno2;
    return -1;
  }
  /* Lock object registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized object registry. */
  if (0 == rmix_objects_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(object registry finalized)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Search for object id in registry. */
  for (index = 0; index < rmix_objects_data.count; index++) {
    if (objectid == rmix_objects_data.objects[index].objectid) {
      break;
    }
  }
  if (index == rmix_objects_data.count) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(unable to find object id in registry)))
    /* Unlock object registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return -1;
  }
  /* Copy object and interface. */
  *object    = rmix_objects_data.objects[index].object;
  *interface = rmix_objects_data.objects[index].interface;
  /* Unlock object registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_objects_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock object registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
