/******************************************************************************
 *
 * Header file for rmix stream (I/O stream) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_STREAM_H
#define RMIX_STREAM_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* Incoming data stream "read" function signature. */
/*
 * Reads from an incoming data stream.
 *
 * incoming = The incoming data stream object.
 * size     = The number of bytes to read.
 * data     = The byte buffer to read into.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
#define RMIX_STREAM_READ(name)\
  int name (void           *incoming,\
            unsigned int    size,\
            void           *data)

/* Outgoing data stream "write" function signature. */
/*
 * Writes into an outgoing data stream.
 *
 * outgoing = The outgoing data stream object.
 * size     = The number of bytes to write.
 * data     = The byte buffer to write from.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
#define RMIX_STREAM_WRITE(name)\
  int name (void           *outgoing,\
            unsigned int    size,\
            const void     *data)

/* Data stream "close" function signature. */
/*
 * Closes a data stream.
 *
 * stream = The data stream object.
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#define RMIX_STREAM_CLOSE(name)\
  int name (void *stream)

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Incoming data stream "read" function data type. */
typedef RMIX_STREAM_READ((*rmix_stream_read_t));

/* Outgoing data stream "write" function data type. */
typedef RMIX_STREAM_WRITE((*rmix_stream_write_t));

/* Outgoing data stream "close" function data type. */
typedef RMIX_STREAM_CLOSE((*rmix_stream_close_t));

/* Incoming data stream data type. */
typedef struct {
  void                *incoming; /* stream object  */
  rmix_stream_read_t   read;     /* read function  */
  rmix_stream_close_t  close;    /* close function */
} rmix_istream_t;

/* Outgoing data stream data type. */
typedef struct {
  void                *outgoing; /* stream object  */
  rmix_stream_write_t  write;    /* write function */
  rmix_stream_close_t  close;    /* close function */
} rmix_ostream_t;

/* Data stream data type. */
typedef struct {
  rmix_istream_t istream; /* incoming data stream. */
  rmix_ostream_t ostream; /* outgoing data stream. */
} rmix_stream_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Creates a data stream.
 *
 * stream   = The data stream (return).
 * incoming = The incoming stream object.
 * read     = The incoming stream read function.
 * iclose   = The incoming stream close function.
 * outgoing = The outgoing stream object.
 * write    = The outgoing stream write function.
 * oclose   = The outgoing stream close function.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_stream_create (rmix_stream_t       **stream,
                        void                 *incoming,
                        rmix_stream_read_t    read,
                        rmix_stream_close_t   iclose,
                        void                 *outgoing,
                        rmix_stream_write_t   write,
                        rmix_stream_close_t   oclose);

/*
 * Destroys a data stream.
 *
 * stream = The data stream (return).
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_stream_destroy (rmix_stream_t **stream);

/*
 * Creates an incoming data stream.
 *
 * istream  = The incoming data stream (return).
 * incoming = The incoming stream object.
 * read     = The incoming stream read function.
 * close    = The incoming stream close function.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_istream_create (rmix_istream_t     **istream,
                         void                *incoming,
                         rmix_stream_read_t   read,
                         rmix_stream_close_t  close);

/*
 * Destroys an incoming data stream.
 *
 * istream = The incoming data stream (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_istream_destroy (rmix_istream_t **istream);

/*
 * Creates an outgoing data stream.
 *
 * ostream  = The outgoing data stream (return).
 * outgoing = The outgoing stream object.
 * write    = The outgoing stream write function.
 * close    = The outgoing stream close function.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_ostream_create (rmix_ostream_t      **ostream,
                         void                 *outgoing,
                         rmix_stream_write_t   write,
                         rmix_stream_close_t   close);

/*
 * Destroys an outgoing data stream.
 *
 * istream = The outgoing data stream (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_ostream_destroy (rmix_ostream_t **ostream);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_STREAM_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
