/******************************************************************************
 *
 * Source file for rmix threadids (thread id registry) module.
 * Copyright (c) 2002-2005 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Thread id registry data type. */
typedef struct {
  pthread_mutex_t mutex; /* mutex            */
  unsigned int    ready; /* initialized flag */
  pthread_key_t   key;   /* thread id key    */
} rmix_threadids_data_t;

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Thread id registry data. */
rmix_threadids_data_t rmix_threadids_data = {
  PTHREAD_MUTEX_INITIALIZER, /* mutex            */
  0                          /* initialized flag */
  /* uninitialized */        /* thread id key    */
};

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Initializes the thread id registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threadids_init ();

/*
 * Finalizes the thread id registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threadids_fini ();

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Initializes the thread id registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threadids_init"
int rmix_threadids_init () {
  int error;
  RMIX_LOG((RMIX_INFO(thread id registry startup)))
  /* Lock thread id registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threadids_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread id registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for initialized thread id registry. */
  if (1 == rmix_threadids_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(thread id registry already initialized)))
    /* Unlock thread id registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threadids_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread id registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Create thread id key. */
  if (0 != (error = pthread_key_create(&rmix_threadids_data.key, NULL))) {
    int error2 = error;
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to create thread id key)))
    errno = error;
    rmix_syserr();
    /* Unlock thread id registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threadids_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread id registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = error2;
    return -1;
  }
  /* Mark thread id registry initialized. */
  rmix_threadids_data.ready = 1;
  /* Unlock thread id registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threadids_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread id registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    /* Mark thread id registry finalized. */
    rmix_threadids_data.ready = 0;
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the thread id registry.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threadids_fini"
int rmix_threadids_fini () {
  int          error;
  RMIX_LOG((RMIX_INFO(thread id registry shutdown)))
  /* Lock thread id registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threadids_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread id registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized thread id registry. */
  if (0 == rmix_threadids_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(thread id registry already finalized)))
    /* Unlock thread id registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threadids_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread id registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Mark thread id registry finalized. */
  rmix_threadids_data.ready = 0;
  /* Delete thread id key. */
  if (0 != (error = pthread_key_delete(rmix_threadids_data.key))) {
    int error2 = error;
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to create thread id key)))
    errno = error;
    rmix_syserr();
    /* Unlock thread id registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threadids_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread id registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = error2;
    return -1;
  }
  /* Unlock thread id registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threadids_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread id registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Returns a unique id for the calling thread from the rmix thread id registry.
 *
 * threadid = The unique id for the calling thread.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threadids_get"
int rmix_threadids_get (unsigned int *threadid) {
  int error;
  /* Lock thread id registry mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threadids_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread id registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for finalized thread id registry. */
  if (0 == rmix_threadids_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(thread id registry finalized)))
    /* Unlock thread id registry mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threadids_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread id registry mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Retrieve thread id. Generate and save it if it doesn't exist yet. */
  while (0 == (*threadid =
               (unsigned int)pthread_getspecific(rmix_threadids_data.key))) {
    /* Generate thread id randomly. */
    *threadid = (unsigned int)random();
    /* Save thread id. */
    if (0 != (error = pthread_setspecific(rmix_threadids_data.key,
                                          (const void*)*threadid))) {
      int error2 = error;
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to save thread id)))
      /* Unlock thread id registry mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_threadids_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock thread id registry mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = error2;
      return -1;
    }
  }
  /* Unlock thread id registry mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threadids_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread id registry mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
