/******************************************************************************
 *
 * Source file for rmix threads (thread pool) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* Main module header file. */
#include "rmix.h"

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* Thread pool data type. */
typedef struct {
  pthread_mutex_t                 mutex;    /* mutex                    */
  unsigned int                    ready;    /* initialized flag         */
  struct {
    pthread_cond_t                exited;   /* exited  thread condition */
    unsigned int                  minimum;  /* minimum thread count     */
    unsigned int                  current;  /* current thread count     */
    unsigned int                  maximum;  /* maximum thread count     */
    unsigned int                  waiting;  /* waiting thread count     */
    unsigned int                  timeout;  /* exiting thread timeout   */
    unsigned int                  size;     /* thread array size        */
    pthread_t                    *threads;  /* thread array             */
    pthread_attr_t                attribs;  /* thread attributes        */
  }                               pool;
  struct {
    pthread_cond_t                enqueued; /* enqueued job condition   */
    pthread_cond_t                dequeued; /* dequeued job condition   */
    unsigned int                  first;    /* first   job index        */
    unsigned int                  current;  /* current job count        */
    unsigned int                  maximum;  /* maximum job count        */
    unsigned int                  timeout;  /* waiting thread timeout   */
    unsigned int                  draining; /* draining flag            */
    unsigned int                  size;     /* job queue size           */
    struct {
      rmix_threads_job_t          function; /* job function             */
      rmix_threads_job_cleanup_t  cleanup;  /* job cleanup function     */
      void                       *argument; /* job argument             */
    }                            *queue;    /* job queue                */
  }                               jobs;
} rmix_threads_data_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Sets the thread pool configuration and initializes the thread pool.
 *
 * pool_minimum = The new minimum thread count.
 * pool_maximum = The new maximum thread count.
 * pool_timeout = The new exiting thread timeout.
 * jobs_maximum = The new maximum job count.
 * jobs_timeout = The new waiting thread timeout.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_init (unsigned int pool_minimum,
                       unsigned int pool_maximum,
                       unsigned int pool_timeout,
                       unsigned int jobs_maximum,
                       unsigned int jobs_timeout);

/*
 * Initializes the thread pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_init2 ();

/*
 * Finalizes the thread pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_threads_fini ();

/*
 * The thread pool thread dequeues jobs from the job queue and executes them.
 *
 * argument = NULL.
 * return   = NULL.
 */
void *rmix_threads_run (void *argument);

/*
 * The thread pool thread exit cleanup.
 *
 * argument = NULL.
 * return   = NULL.
 */
void rmix_threads_exit (void *argument);

/*
 * The thread pool mutex unlock cleanup.
 *
 * argument = NULL.
 * return   = NULL.
 */
void rmix_threads_unlock (void *argument);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* Thread pool data. */
rmix_threads_data_t rmix_threads_data = {
  PTHREAD_MUTEX_INITIALIZER,   /* mutex                    */
  0,                           /* initialized flag         */
  {
    PTHREAD_COND_INITIALIZER,  /* exited  thread condition */
    RMIX_THREADS_POOL_MINIMUM, /* minimum thread count     */
    0,                         /* current thread count     */
    RMIX_THREADS_POOL_MAXIMUM, /* maximum thread count     */
    0,                         /* waiting thread count     */
    RMIX_THREADS_POOL_TIMEOUT, /* exiting thread timeout   */
    0,                         /* thread array size        */
    NULL                       /* thread array             */
    /* not initialized */      /* thread attributes        */
  },
  {
    PTHREAD_COND_INITIALIZER,  /* enqueued job condition   */
    PTHREAD_COND_INITIALIZER,  /* dequeued job condition   */
    0,                         /* first   job index        */
    0,                         /* current job count        */
    RMIX_THREADS_JOBS_MAXIMUM, /* maximum job count        */
    RMIX_THREADS_JOBS_TIMEOUT, /* waiting thread timeout   */
    0,                         /* draining flag            */
    0,                         /* job queue size           */
    NULL                       /* job queue                */
  }
};

/******************************************************************************
 *
 * Functions
 *
 ******************************************************************************/

/*
 * Sets the minimum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * minimum = The new minimum thread count.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_pool_minimum_set"
int rmix_threads_pool_minimum_set (unsigned int minimum) {
  int error;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check minimum parameter. */
  if (minimum > rmix_threads_data.pool.maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(minimum parameter is more than maximum)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  /* Check for initialized thread pool. */
  if (1 == rmix_threads_data.ready) {
    /* Check for a decrease of the minimum thread count of the thread pool. */
    if (rmix_threads_data.pool.minimum > minimum) {
      /* Broadcast enqueued job to decrease the current to
         match the minimum thread count of the thread pool. */
      if (0 != (error =
                pthread_cond_broadcast(&rmix_threads_data.jobs.enqueued))) {
        int error2 = error;
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to broadcast enqueued job)))
        errno = error;
        rmix_syserr();
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = error2;
        return -1;
      }
    /* Check for an increase of the minimum thread count of the thread pool. */
    } else if (rmix_threads_data.pool.minimum < minimum) {
      /* Increase the current to match the
         minimum thread count of the thread pool. */
      for (;rmix_threads_data.pool.current < minimum;
            rmix_threads_data.pool.current++) {
        /* Create thread. */
        if (0 != (error = pthread_create(&rmix_threads_data.pool.threads
                                         [rmix_threads_data.pool.current],
                                         &rmix_threads_data.pool.attribs,
                                          rmix_threads_run, NULL))) {
          int error2 = error;
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to create thread)))
          /* Unlock thread pool mutex. */
          if (0 != (error =
                    pthread_mutex_unlock(&rmix_threads_data.mutex))) {
            errno = error;
            RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
            errno = error;
            rmix_syserr();
          }
          errno = error2;
          return -1;
        }
      }
    }
  }
  /* Set the minimum thread count of the thread pool. */
  rmix_threads_data.pool.minimum = minimum;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Sets the maximum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * maximum = The new maximum thread count.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_pool_maximum_set"
int rmix_threads_pool_maximum_set (unsigned int maximum) {
  int           error;
  pthread_t    *threads;
  unsigned int  current;
  /* Check maximum parameter. */
  if (0 == maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum parameter is zero)))
    return -1;
  }
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check maximum parameter. */
  if (maximum < rmix_threads_data.pool.minimum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum parameter is less than minimum)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  /* Check for initialized thread pool. */
  if (1 == rmix_threads_data.ready) {
    /* Check for a decrease of the maximum thread count of the thread pool. */
    if (rmix_threads_data.pool.maximum > maximum) {
      /* Broadcast enqueued job to decrease the current to
         match the maximum thread count of the thread pool. */
      if (0 != (error = 
                pthread_cond_broadcast(&rmix_threads_data.jobs.enqueued))) {
        int error2 = error;
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to broadcast enqueued job)))
        errno = error;
        rmix_syserr();
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = error2;
        return -1;
      }
    /* Check for an increase of the maximum thread count of the thread pool. */
    } else if (rmix_threads_data.pool.maximum < maximum) {
      /* Reallocate the thread pool. */
      if (NULL == (threads =
                   (pthread_t*)realloc(rmix_threads_data.pool.threads,
                                       (rmix_threads_data.pool.size +
                                        (maximum - 
                                         rmix_threads_data.pool.maximum)) *
                                       sizeof(pthread_t)))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to reallocate the thread pool)))
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = errno2;
        return -1;
      }
      rmix_threads_data.pool.threads  = threads;
      rmix_threads_data.pool.size    += (maximum - 
                                        rmix_threads_data.pool.maximum);
      /* Calculate the appropiate current thread count to
         match the maximum thread count of the thread pool. */
      current = rmix_threads_data.pool.current;
      if (rmix_threads_data.jobs.current < maximum - 
                                          rmix_threads_data.pool.maximum) {
        current += rmix_threads_data.jobs.current;
      } else {
        current += (maximum - rmix_threads_data.pool.current);
      }
      /* Create the appropiate current thread count to
         match the maximum thread count of the thread pool. */
      for (;rmix_threads_data.pool.current < current;
            rmix_threads_data.pool.current++) {
        /* Create thread. */
        if (0 != (error = pthread_create(&rmix_threads_data.pool.threads
                                         [rmix_threads_data.pool.current],
                                         &rmix_threads_data.pool.attribs,
                                          rmix_threads_run, NULL))) {
          int error2 = error;
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to create thread)))
          /* Unlock thread pool mutex. */
          if (0 != (error =
                    pthread_mutex_unlock(&rmix_threads_data.mutex))) {
            errno = error;
            RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
            errno = error;
            rmix_syserr();
          }
          errno = error2;
          return -1;
        }
      }
    }
  }
  /* Set the maximum thread count of the thread pool. */
  rmix_threads_data.pool.maximum = maximum;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Sets the exiting thread timeout of the thread pool.
 * This function may be called before initializing.
 *
 * timeout = The new exiting thread timeout.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_pool_timeout_set"
int rmix_threads_pool_timeout_set (unsigned int timeout) {
  int error;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Set the exiting thread timeout of the thread pool. */
  rmix_threads_data.pool.timeout = timeout;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Sets the maximum job count of the thread pools job queue.
 * This function may be called before initializing.
 *
 * maximum = The new maximum job count.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_jobs_maximum_set"
int rmix_threads_jobs_maximum_set (unsigned int maximum) {
  int   error;
  void *queue;
  /* Check maximum parameter. */
  if (0 == maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum parameter is zero)))
    return -1;
  }
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for initialized thread pool. */
  if (1 == rmix_threads_data.ready) {
    /* Check for an increase of the size of the thread pools job queue. */
    if (rmix_threads_data.jobs.size < maximum) {
      /* Reallocate the job queue of the thread pool. */
      if (NULL == (queue =
                   realloc(rmix_threads_data.jobs.queue, maximum *
                           sizeof(rmix_threads_data.jobs.queue[0])))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to reallocate the job queue)))
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = errno2;
        return -1;
      }
      rmix_threads_data.jobs.queue = queue;
      /* Check if jobs in job queue wrap around. */
      if (rmix_threads_data.jobs.first +
          rmix_threads_data.jobs.current >
          rmix_threads_data.jobs.size) {
        /* Rewrap job queue at new size. */
        memcpy(rmix_threads_data.jobs.queue + 
               rmix_threads_data.jobs.size,
               rmix_threads_data.jobs.queue,
               (maximum - rmix_threads_data.jobs.size) *
               sizeof(rmix_threads_data.jobs.queue[0]));
        if (rmix_threads_data.jobs.first >
            maximum - rmix_threads_data.jobs.size) {
          memcpy(rmix_threads_data.jobs.queue,
                 rmix_threads_data.jobs.queue +
                 (maximum - rmix_threads_data.jobs.size),
                 (rmix_threads_data.jobs.first -
                  (maximum - rmix_threads_data.jobs.size)) *
                 sizeof(rmix_threads_data.jobs.queue[0]));
        }
      }
      rmix_threads_data.jobs.size = maximum;
    }
    /* Check for an increase of the maximum job count of the thread pool. */
    if (rmix_threads_data.jobs.maximum < maximum) {
      /* Broadcast dequeued job to wakeup waiting threads. */
      if (0 != (error = 
                pthread_cond_broadcast(&rmix_threads_data.jobs.dequeued))) {
        int error2 = error;
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to broadcast dequeued job)))
        errno = error;
        rmix_syserr();
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = error2;
        return -1;
      }
    }
  }
  /* Set the maximum job count of the thread pool. */
  rmix_threads_data.jobs.maximum = maximum;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Sets the waiting thread timeout of the thread pools job queue.
 * This function may be called before initializing.
 *
 * timeout = The new waiting thread timeout.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_jobs_timeout_set"
int rmix_threads_jobs_timeout_set (unsigned int timeout) {
  int error;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Set the waiting thread timeout of the thread pool. */
  rmix_threads_data.jobs.timeout = timeout;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Increases/Decreases the minimum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * minimum = The minimum thread count increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_pool_minimum_add"
int rmix_threads_pool_minimum_add (int minimum) {
  int error;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Calculate the new minimum thread count of the thread pool. */
  minimum += rmix_threads_data.pool.minimum;
  /* Check the new minimum thread count of the thread pool. */
  if (0 > minimum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(minimum thread count is negative)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  if (minimum > rmix_threads_data.pool.maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(minimum thread count is more than maximum)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  /* Check for initialized thread pool. */
  if (1 == rmix_threads_data.ready) {
    /* Check for a decrease of the minimum thread count of the thread pool. */
    if (rmix_threads_data.pool.minimum > minimum) {
      /* Broadcast enqueued job to decrease the current to
         match the minimum thread count of the thread pool. */
      if (0 != (error = 
                pthread_cond_broadcast(&rmix_threads_data.jobs.enqueued))) {
        int error2 = error;
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to broadcast enqueued job)))
        errno = error;
        rmix_syserr();
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = error2;
        return -1;
      }
    /* Check for an increase of the minimum thread count of the thread pool. */
    } else if (rmix_threads_data.pool.minimum < minimum) {
      /* Increase the current to match the
         minimum thread count of the thread pool. */
      for (;rmix_threads_data.pool.current < minimum;
            rmix_threads_data.pool.current++) {
        /* Create thread. */
        if (0 != (error = pthread_create(&rmix_threads_data.pool.threads
                                         [rmix_threads_data.pool.current],
                                         &rmix_threads_data.pool.attribs,
                                          rmix_threads_run, NULL))) {
          int error2 = error;
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to create thread)))
          /* Unlock thread pool mutex. */
          if (0 != (error = 
                    pthread_mutex_unlock(&rmix_threads_data.mutex))) {
            errno = error;
            RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
            errno = error;
            rmix_syserr();
          }
          errno = error2;
          return -1;
        }
      }
    }
  }
  /* Set the minimum thread count of the thread pool. */
  rmix_threads_data.pool.minimum = minimum;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Increases/Decreases the maximum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * maximum = The maximum thread count increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_pool_maximum_add"
int rmix_threads_pool_maximum_add (int maximum) {
  int           error;
  pthread_t    *threads;
  unsigned int  current;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Calculate the new maximum thread count of the thread pool. */
  maximum += rmix_threads_data.pool.maximum;
  /* Check new maximum thread count of the thread pool. */
  if (0 == maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum thread count is zero)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  if (maximum < rmix_threads_data.pool.minimum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum thread count is less than minimum)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  /* Check for initialized thread pool. */
  if (1 == rmix_threads_data.ready) {
    /* Check for a decrease of the maximum thread count of the thread pool. */
    if (rmix_threads_data.pool.maximum > maximum) {
      /* Broadcast enqueued job to decrease the current to
         match the maximum thread count of the thread pool. */
      if (0 != (error = 
                pthread_cond_broadcast(&rmix_threads_data.jobs.enqueued))) {
        int error2 = error;
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to broadcast enqueued job)))
        errno = error;
        rmix_syserr();
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = error2;
        return -1;
      }
    /* Check for an increase of the maximum thread count of the thread pool. */
    } else if (rmix_threads_data.pool.maximum < maximum) {
      /* Reallocate the thread pool. */
      if (NULL == (threads =
                   (pthread_t*)realloc(rmix_threads_data.pool.threads,
                                       (rmix_threads_data.pool.size +
                                        (maximum - 
                                         rmix_threads_data.pool.maximum)) *
                                       sizeof(pthread_t)))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to reallocate the thread pool)))
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = errno2;
        return -1;
      }
      rmix_threads_data.pool.threads  = threads;
      rmix_threads_data.pool.size    += (maximum - 
                                        rmix_threads_data.pool.maximum);
      /* Calculate the appropiate current thread count to
         match the maximum thread count of the thread pool. */
      current = rmix_threads_data.pool.current;
      if (rmix_threads_data.jobs.current < maximum - 
                                          rmix_threads_data.pool.maximum) {
        current += rmix_threads_data.jobs.current;
      } else {
        current += (maximum - rmix_threads_data.pool.current);
      }
      /* Create the appropiate current thread count to
         match the maximum thread count of the thread pool. */
      for (;rmix_threads_data.pool.current < current;
            rmix_threads_data.pool.current++) {
        /* Create thread. */
        if (0 != (error = pthread_create(&rmix_threads_data.pool.threads
                                         [rmix_threads_data.pool.current],
                                         &rmix_threads_data.pool.attribs,
                                          rmix_threads_run, NULL))) {
          int error2 = error;
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to create thread)))
          /* Unlock thread pool mutex. */
          if (0 != (error =
                    pthread_mutex_unlock(&rmix_threads_data.mutex))) {
            errno = error;
            RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
            errno = error;
            rmix_syserr();
          }
          errno = error2;
          return -1;
        }
      }
    }
  }
  /* Set the maximum thread count of the thread pool. */
  rmix_threads_data.pool.maximum = maximum;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Increases/Decreases the exiting thread timeout of the thread pool.
 * This function may be called before initializing.
 *
 * timeout = The exiting thread timeout increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_pool_timeout_add"
int rmix_threads_pool_timeout_add (int timeout) {
  int error;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Calculate the new exiting thread timeout of the thread pool. */
  timeout += rmix_threads_data.pool.timeout;
  /* Check new exiting thread timeout of the thread pool. */
  if (0 > timeout) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(exiting thread timeout is negative)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  /* Set the exiting thread timeout of the thread pool. */
  rmix_threads_data.pool.timeout = timeout;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Increases/Decreases the maximum job count of the thread pools job queue.
 * This function may be called before initializing.
 *
 * maximum = The maximum job count increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_jobs_maximum_add"
int rmix_threads_jobs_maximum_add (int maximum) {
  int   error;
  void *queue;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for initialized thread pool. */
  if (1 == rmix_threads_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(thread pool already initialized)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Calculate the new maximum job count of the thread pool. */
  maximum += rmix_threads_data.jobs.maximum;
  /* Check new maximum job count of the thread pool. */
  if (0 == maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum job count is zero)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  if (0 > maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum job count is negative)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  /* Check for initialized thread pool. */
  if (1 == rmix_threads_data.ready) {
    /* Check for an increase of the size of the thread pools job queue. */
    if (rmix_threads_data.jobs.size < maximum) {
      /* Reallocate the job queue of the thread pool. */
      if (NULL == (queue =
                   realloc(rmix_threads_data.jobs.queue, maximum *
                           sizeof(rmix_threads_data.jobs.queue[0])))) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to reallocate the job queue)))
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = errno2;
        return -1;
      }
      rmix_threads_data.jobs.queue = queue;
      /* Check if jobs in job queue wrap around. */
      if (rmix_threads_data.jobs.first +
          rmix_threads_data.jobs.current >
          rmix_threads_data.jobs.size) {
        /* Rewrap job queue at new size. */
        memcpy(rmix_threads_data.jobs.queue +
               rmix_threads_data.jobs.size,
               rmix_threads_data.jobs.queue,
               (maximum - rmix_threads_data.jobs.size) *
               sizeof(rmix_threads_data.jobs.queue[0]));
        if (rmix_threads_data.jobs.first >
            maximum - rmix_threads_data.jobs.size) {
          memcpy(rmix_threads_data.jobs.queue,
                 rmix_threads_data.jobs.queue +
                 (maximum - rmix_threads_data.jobs.size),
                 (rmix_threads_data.jobs.first -
                  (maximum - rmix_threads_data.jobs.size)) *
                 sizeof(rmix_threads_data.jobs.queue[0]));
        }
      }
      rmix_threads_data.jobs.size = maximum;
    }
    /* Check for an increase of the maximum job count of the thread pool. */
    if (rmix_threads_data.jobs.maximum < maximum) {
      /* Broadcast dequeued job to wakeup waiting threads. */
      if (0 != (error = 
                pthread_cond_broadcast(&rmix_threads_data.jobs.dequeued))) {
        int error2 = error;
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to broadcast dequeued job)))
        errno = error;
        rmix_syserr();
        /* Unlock thread pool mutex. */
        if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
          errno = error;
          rmix_syserr();
        }
        errno = error2;
        return -1;
      }
    }
  }
  /* Set the maximum job count of the thread pool. */
  rmix_threads_data.jobs.maximum = maximum;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Increases/Decreases the waiting thread timeout of the thread pools job queue.
 * This function may be called before initializing.
 *
 * timeout = The waiting thread timeout increment/decrement.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_jobs_timeout_add"
int rmix_threads_jobs_timeout_add (int timeout) {
  int error;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Calculate the new waiting thread timeout of the thread pool. */
  timeout += rmix_threads_data.jobs.timeout;
  /* Check new waiting thread timeout of the thread pool. */
  if (0 > timeout) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(waiting thread timeout is negative)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EINVAL;
    return -1;
  }
  /* Set the waiting thread timeout of the thread pool. */
  rmix_threads_data.jobs.timeout = timeout;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Gets the minimum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * minimum = The minimum thread count (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_pool_minimum_get"
int rmix_threads_pool_minimum_get (unsigned int *minimum) {
  int error;
#ifdef DEBUG
  /* Check minimum parameter. */
  if (NULL == minimum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(minimum parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Copy the minimum thread count of the thread pool. */
  *minimum = rmix_threads_data.pool.minimum;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Gets the maximum thread count of the thread pool.
 * This function may be called before initializing.
 *
 * maximum = The maximum thread count (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_pool_maximum_get"
int rmix_threads_pool_maximum_get (unsigned int *maximum) {
  int error;
#ifdef DEBUG
  /* Check maximum parameter. */
  if (NULL == maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Copy the maximum thread count of the thread pool. */
  *maximum = rmix_threads_data.pool.maximum;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Gets the exiting thread timeout of the thread pool.
 * This function may be called before initializing.
 *
 * timeout = The exiting thread timeout (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_pool_timeout_get"
int rmix_threads_pool_timeout_get (unsigned int *timeout) {
  int error;
#ifdef DEBUG
  /* Check timeout parameter. */
  if (NULL == timeout) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(timeout parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Copy the exiting thread timeout of the thread pool. */
  *timeout = rmix_threads_data.pool.timeout;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Gets the maximum job count of the thread pools job queue.
 * This function may be called before initializing.
 *
 * maximum = The maximum job count (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_jobs_maximum_get"
int rmix_threads_jobs_maximum_get (unsigned int *maximum){
  int error;
#ifdef DEBUG
  /* Check maximum parameter. */
  if (NULL == maximum) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(maximum parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Copy the maximum job count of the thread pool.. */
  *maximum = rmix_threads_data.jobs.maximum;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Gets the waiting thread timeout of the thread pools job queue.
 * This function may be called before initializing.
 *
 * timeout = The waiting thread timeout (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_jobs_timeout_get"
int rmix_threads_jobs_timeout_get (unsigned int *timeout) {
  int error;
#ifdef DEBUG
  /* Check timeout parameter. */
  if (NULL == timeout) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(timeout parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Copy the waiting thread timeout of the thread pool. */
  *timeout = rmix_threads_data.jobs.timeout;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Sets the thread pool configuration and initializes the thread pool.
 *
 * pool_minimum = The new minimum thread count.
 * pool_maximum = The new maximum thread count.
 * pool_timeout = The new exiting thread timeout.
 * jobs_maximum = The new maximum job count.
 * jobs_timeout = The new waiting thread timeout.
 * return       = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_init"
int rmix_threads_init (unsigned int pool_minimum,
                       unsigned int pool_maximum,
                       unsigned int pool_timeout,
                       unsigned int jobs_maximum,
                       unsigned int jobs_timeout) {
  unsigned int pool_maximum_old;
  /* Get the maximum thread count of the thread pool. */
  if (0 != rmix_threads_pool_minimum_get(&pool_maximum_old)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get the maximum thread count)))
    errno = errno2;
    return -1;
  }
  if (pool_minimum > pool_maximum_old) {
    /* Set the maximum thread count of the thread pool. */
    if (0 != rmix_threads_pool_maximum_set(pool_maximum)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to set the maximum thread count)))
      errno = errno2;
      return -1;
    }
    /* Set the minimum thread count of the thread pool. */
    if (0 != rmix_threads_pool_minimum_set(pool_minimum)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to set the minimum thread count)))
      errno = errno2;
      return -1;
    }
  } else {
    /* Set the minimum thread count of the thread pool. */
    if (0 != rmix_threads_pool_minimum_set(pool_minimum)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to set the minimum thread count)))
      errno = errno2;
      return -1;
    }
    /* Set the maximum thread count of the thread pool. */
    if (0 != rmix_threads_pool_maximum_set(pool_maximum)) {
      int errno2 = errno;
      RMIX_LOG((RMIX_WARN(unable to set the maximum thread count)))
      errno = errno2;
      return -1;
    }
  }
  /* Set the exiting thread timeout of the thread pool. */
  if (0 != rmix_threads_pool_timeout_set(pool_timeout)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to set the exiting thread timeout)))
    errno = errno2;
    return -1;
  }
  /* Set the maximum job count of the thread pools job queue. */
  if (0 != rmix_threads_jobs_maximum_set(jobs_maximum)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to set the maximum job count)))
    errno = errno2;
    return -1;
  }
  /* Set the waiting thread timeout of the thread pools job queue. */
  if (0 != rmix_threads_jobs_timeout_set(jobs_timeout)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to set the waiting thread timeout)))
    errno = errno2;
    return -1;
  }
  /* Initialize the thread pool. */
  if (0 != rmix_threads_init2()) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to initialize the thread pool)))
    errno = errno2;
    return -1;
  }
  return 0;
}

/*
 * Initializes the thread pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_init2"
int rmix_threads_init2 () {
  int error;
  RMIX_LOG((RMIX_INFO(thread pool startup)))
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Check for initialized thread pool. */
  if (1 == rmix_threads_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(thread pool already initialized)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Initialize thread attributes. */
  if (0 != (error = pthread_attr_init(&rmix_threads_data.pool.attribs))) {
    int error2 = error;
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to initialize thread attributes)))
    errno = error;
    rmix_syserr();
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = error2;
    return -1;
  }
  /* Set thread attributes detach state. */
  if (0 != (error =
            pthread_attr_setdetachstate(&rmix_threads_data.pool.attribs,
                                        PTHREAD_CREATE_DETACHED))) {
    int error2 = error;
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to set thread attributes detach state)))
    errno = error;
    rmix_syserr();
    /* Finalize thread attributes. */
    if (0 != (error =
              pthread_attr_destroy(&rmix_threads_data.pool.attribs))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to finalize thread attributes)))
      errno = error;
      rmix_syserr();
    }
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = error2;
    return -1;
  }
  /* Allocate the thread pool. */
  if (NULL == (rmix_threads_data.pool.threads =
               (pthread_t*)malloc(rmix_threads_data.pool.maximum *
                           sizeof(rmix_threads_data.pool.threads[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate the thread pool)))
    /* Finalize thread attributes. */
    if (0 != (error =
              pthread_attr_destroy(&rmix_threads_data.pool.attribs))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to finalize thread attributes)))
      errno = error;
      rmix_syserr();
    }
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  rmix_threads_data.pool.size = rmix_threads_data.pool.maximum;
  /* Allocate the job queue. */
  if (NULL == (rmix_threads_data.jobs.queue =
               malloc(rmix_threads_data.jobs.maximum *
               sizeof(rmix_threads_data.jobs.queue[0])))) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to allocate the job queue)))
    /* Deallocate the thread pool. */
    free(rmix_threads_data.pool.threads);
    rmix_threads_data.pool.threads = NULL;
    rmix_threads_data.pool.size    = 0; 
    /* Finalize thread attributes. */
    if (0 != (error =
              pthread_attr_destroy(&rmix_threads_data.pool.attribs))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to finalize thread attributes)))
      errno = error;
      rmix_syserr();
    }
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Initialize the job queue. */
  rmix_threads_data.jobs.size     = rmix_threads_data.jobs.maximum;
  rmix_threads_data.jobs.draining = 0;
  rmix_threads_data.jobs.current  = 0;
  rmix_threads_data.jobs.first    = 0;
  /* Create the minimum thread count increase in the thread pool. */
  for (rmix_threads_data.pool.current = 0,
       rmix_threads_data.pool.waiting = 0;
       rmix_threads_data.pool.current < rmix_threads_data.pool.minimum;
       rmix_threads_data.pool.current++) {
    if (0 != (error = pthread_create(&rmix_threads_data.pool.threads
                                     [rmix_threads_data.pool.current],
                                     &rmix_threads_data.pool.attribs,
                                      rmix_threads_run, NULL))) {
      int error2 = error;
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to create thread)))
      /* Cancel all threads in the thread pool. */
      while (0 != rmix_threads_data.pool.current) {
        rmix_threads_data.pool.current--;
        if (0 != (error =
                  pthread_cancel( rmix_threads_data.pool.threads
                                 [rmix_threads_data.pool.current]))) {
          errno = error;
          RMIX_LOG((RMIX_WARN(unable to cancel thread)))
          errno = error;
          rmix_syserr();
        }
      }
      /* Deallocate the job queue. */
      free(rmix_threads_data.jobs.queue);
      rmix_threads_data.jobs.queue = NULL;
      rmix_threads_data.jobs.size  = 0; 
      /* Deallocate the thread pool. */
      free(rmix_threads_data.pool.threads);
      rmix_threads_data.pool.threads = NULL;
      rmix_threads_data.pool.size    = 0; 
      /* Finalize thread attributes. */
      if (0 != (error =
                pthread_attr_destroy(&rmix_threads_data.pool.attribs))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to finalize thread attributes)))
        errno = error;
        rmix_syserr();
      }
      /* Unlock thread pool mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = error2;
      return -1;
    }
  }
  /* Mark thread pool initialized. */
  rmix_threads_data.ready = 1;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    /* Mark thread pool finalized. */
    rmix_threads_data.ready = 0;
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * Finalizes the thread pool.
 *
 * return = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_fini"
int rmix_threads_fini () {
  int             error, result, errno3;
  unsigned int    index;
  struct timeval  timeofday;
  struct timespec timeout;
  RMIX_LOG((RMIX_INFO(thread pool shutdown)))
  result = 0;
  errno3 = 0;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Push unlock thread pool mutex to pthread cleanup stack. */
  pthread_cleanup_push(rmix_threads_unlock, NULL);
  /* Check for finalized thread pool. */
  if (0 == rmix_threads_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(thread pool already finalized)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Get time of day for exiting thread timeout. */
  if (0 != gettimeofday(&timeofday, NULL)) {
    int errno2 = errno;
    RMIX_LOG((RMIX_WARN(unable to get time of day)))
    errno = errno2;
    rmix_syserr();
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = errno2;
    return -1;
  }
  /* Calculate absolute time of exiting thread timeout. */
  timeout.tv_sec  = timeofday.tv_sec  + rmix_threads_data.pool.timeout;
  timeout.tv_nsec = timeofday.tv_usec * 1000;
  /* Drain the thread pool. */
  rmix_threads_data.jobs.draining = 1;
  /* Wait for all threads in the thread
     pool to exit with exiting thread timeout. */
  while (0 != rmix_threads_data.pool.current) {
    error = pthread_cond_timedwait(&rmix_threads_data.pool.exited, 
                                   &rmix_threads_data.mutex, &timeout);
    /* Check if exiting thread timeout was reached. */
    if (ETIMEDOUT == error) {
      break;
    } else if (0 != error) {
      int error2 = error;
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to wait for thread exits)))
      errno = error;
      rmix_syserr();
      /* Unlock thread pool mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
        rmix_syserr();
      }
      errno = error2;
      return -1;
    }
  }
  /* Cancel all threads in the thread pool. */
  for (index = 0; index < rmix_threads_data.pool.current; index++) {
    if (0 != (error =
              pthread_cancel(rmix_threads_data.pool.threads[index]))) {
      int error2 = error;
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to cancel thread)))
      errno = error;
      rmix_syserr();
      /* Unlock thread pool mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = error2;
      return -1;
    }
  }
  /* Wait for all threads in the thread
     pool to exit with exiting thread timeout. */
  while (0 != rmix_threads_data.pool.current) {
    error = pthread_cond_timedwait(&rmix_threads_data.pool.exited, 
                                   &rmix_threads_data.mutex, &timeout);
    /* Check if exiting thread timeout was reached. */
    if (ETIMEDOUT == error) {
      break;
    } else if (0 != error) {
      int error2 = error;
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to wait for thread exits)))
      errno = error;
      rmix_syserr();
      /* Unlock thread pool mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = error2;
      return -1;
    }
  }
  /* Cleanup all queued jobs. */
  for (index = 0; index < rmix_threads_data.jobs.size; index++) {
    /* Check for job cleanup function. */
    if (NULL != rmix_threads_data.jobs.queue[index].cleanup) {
      /* Cleanup queued job. */
      rmix_threads_data.jobs.queue[index].cleanup(
        rmix_threads_data.jobs.queue[index].argument);
    }
  }
  /* Deallocate the job queue. */
  free(rmix_threads_data.jobs.queue);
  rmix_threads_data.jobs.queue = NULL;
  rmix_threads_data.jobs.size  = 0;
  /* Deallocate the thread pool. */
  if (NULL != rmix_threads_data.pool.threads) {
    free(rmix_threads_data.pool.threads);
    rmix_threads_data.pool.threads = NULL;
  }
  rmix_threads_data.pool.size = 0;
  /* Finalize thread attributes. */
  if (0 != (error =
            pthread_attr_destroy(&rmix_threads_data.pool.attribs))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to finalize thread attributes)))
    errno = error;
    rmix_syserr();
    errno3 = error;
    result = -1;
  }
  /* Mark thread pool finalized. */
  rmix_threads_data.ready         = 0; 
  rmix_threads_data.jobs.draining = 0;
  /* Pop unlock thread pool mutex from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  if (0 != errno3) {
    errno = errno3;
  }
  return result;
}

/*
 * Submits a job to the thread pool by enqueueing it into the job queue.
 *
 * function = The job function.
 * cleanup  = The job cleanup function.
 * argument = The job argument (returns NULL if job was enqueued).
 * return   = 0 on success or -1 on any error with errno set appropriately.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_submit"
int rmix_threads_submit (rmix_threads_job_t           function,
                         rmix_threads_job_cleanup_t   cleanup,
                         void                       **argument) {
  int          error;
  unsigned int index;
#ifdef DEBUG
  /* Check function parameter. */
  if (NULL == function) {
    errno = EINVAL;
    RMIX_LOG((RMIX_WARN(function parameter is null)))
    errno = EINVAL;
    return -1;
  }
#endif /* DEBUG */
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  /* Push unlock thread pool mutex to pthread cleanup stack. */
  pthread_cleanup_push(rmix_threads_unlock, NULL);
  /* Check for draining job queue. */
  if (1 == rmix_threads_data.jobs.draining) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(thread pool draining)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Check for finalized thread pool. */
  if (0 == rmix_threads_data.ready) {
    errno = EACCES;
    RMIX_LOG((RMIX_WARN(thread pool finalized)))
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = EACCES;
    return -1;
  }
  /* Check if no thread is waiting and maximum
     thread count in thread pool is not reached. */
  if ((0                             == rmix_threads_data.pool.waiting)&&
      (rmix_threads_data.pool.current <
       rmix_threads_data.pool.maximum)) {
    /* Create thread. */
    if (0 != (error = pthread_create(&rmix_threads_data.pool.threads
                                     [rmix_threads_data.pool.current],
                                     &rmix_threads_data.pool.attribs,
                                      rmix_threads_run, NULL))) {
      int error2 = error;
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to create thread)))
      errno = error;
      rmix_syserr();
      /* Unlock thread pool mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = error2;
      return -1;
    }
    rmix_threads_data.pool.current++;
  }
  /* Check for space in job queue. */
  while (rmix_threads_data.jobs.current >=
         rmix_threads_data.jobs.maximum) {
    /* Wait for space in job queue. */
    if (0 != (error = pthread_cond_wait(&rmix_threads_data.jobs.dequeued,
                                        &rmix_threads_data.mutex))) {
      int error2 = error;
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to wait for space in job queue)))
      errno = error;
      rmix_syserr();
      /* Unlock thread pool mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = error2;
      return -1;
    }
    /* Check for finalized thread pool. */
    if (0 == rmix_threads_data.ready) {
      errno = EACCES;
      RMIX_LOG((RMIX_WARN(thread pool finalized)))
      /* Unlock thread pool mutex. */
      if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
        errno = error;
        rmix_syserr();
      }
      errno = EACCES;
      return -1;
    }
  }
  /* Enqueue new job into job queue. */
  index = (rmix_threads_data.jobs.first +
           rmix_threads_data.jobs.current) %
           rmix_threads_data.jobs.size;
  rmix_threads_data.jobs.queue[index].function = function;
  rmix_threads_data.jobs.queue[index].cleanup  = cleanup;
  if (NULL != argument) {
    rmix_threads_data.jobs.queue[index].argument = *argument;
    *argument = NULL;
  }
  rmix_threads_data.jobs.current++;
  /* Signal enqueued job. */
  if (0 != (error =
            pthread_cond_signal(&rmix_threads_data.jobs.enqueued))) {
    int error2 = error;
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to signal enqueued job)))
    errno = error;
    rmix_syserr();
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = error2;
    return -1;
  }
  /* Pop unlock thread pool mutex from pthread cleanup stack. */
  pthread_cleanup_pop(0);
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return -1;
  }
  return 0;
}

/*
 * The thread pool thread dequeues jobs from the job queue and executes them.
 *
 * argument = NULL.
 * return   = NULL.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_run"
void *rmix_threads_run (void *argument) {
  int                         error;
  void                       *queue;
  unsigned int                index;
  struct timeval              timeofday;
  struct timespec             timeout;
  rmix_threads_job_t          function;
  rmix_threads_job_cleanup_t  cleanup;
  /* Push exit thread to pthread cleanup stack. */
  pthread_cleanup_push(rmix_threads_exit, NULL);
  while (1) {
    /* Lock thread pool mutex. */
    if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
      errno = error;
      rmix_syserr();
      errno = error;
      pthread_exit(NULL);
    }
    /* Push unlock thread pool mutex to pthread cleanup stack. */
    pthread_cleanup_push(rmix_threads_unlock, NULL);
    /* Check for thread cancellation. */
    pthread_testcancel();
    /* Check for finalized thread pool. */
    if (0 == rmix_threads_data.ready) {
      /* Exit thread. */
      pthread_exit(NULL);
    }
    /* Check if current thread count in thread pool is greater than maximum. */
    if (rmix_threads_data.pool.current >
        rmix_threads_data.pool.maximum) {
      /* Exit thread. */
      pthread_exit(NULL);
    }
    /* Check for zero job count in job queue. */
    while (0 == rmix_threads_data.jobs.current) {
      /* Check for draining job queue. */
      if (1 == rmix_threads_data.jobs.draining) {
        /* Exit thread. */
        pthread_exit(NULL);
      }
      /* Get time of day for waiting thread timeout. */
      if (0 != gettimeofday(&timeofday, NULL)) {
        int errno2 = errno;
        RMIX_LOG((RMIX_WARN(unable to get time of day)))
        errno = errno2;
        rmix_syserr();
        errno = errno2;
        pthread_exit(NULL);
      }
      /* Calculate absolute time of waiting thread timeout. */
      timeout.tv_sec  = timeofday.tv_sec  + rmix_threads_data.jobs.timeout;
      timeout.tv_nsec = timeofday.tv_usec * 1000;
      /* Wait for new job in job queue with waiting thread timeout. */
      rmix_threads_data.pool.waiting++;
      error = pthread_cond_timedwait(&rmix_threads_data.jobs.enqueued, 
                                     &rmix_threads_data.mutex, &timeout);
      /* Check for finalized thread pool. */
      if (0 == rmix_threads_data.ready) {
        /* Exit thread. */
        pthread_exit(NULL);
      }
      rmix_threads_data.pool.waiting--;
      /* Check if waiting thread timeout was reached. */
      if (ETIMEDOUT == error) {
        /* Check if current thread count in
           thread pool is greater than minimum. */
        if (rmix_threads_data.pool.current >
            rmix_threads_data.pool.minimum) {
          /* Exit thread. */
          pthread_exit(NULL);
        }
      } else if (0 != error) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to wait for new job in job queue)))
        errno = error;
        rmix_syserr();
        errno = error;
        pthread_exit(NULL);
      }
    }
    /* Dequeue new job from job queue. */
    function = rmix_threads_data.jobs.queue[
                rmix_threads_data.jobs.first].function;
    cleanup  = rmix_threads_data.jobs.queue[
                rmix_threads_data.jobs.first].cleanup;
    argument = rmix_threads_data.jobs.queue[
                rmix_threads_data.jobs.first].argument;
    rmix_threads_data.jobs.current--;
    rmix_threads_data.jobs.first = (rmix_threads_data.jobs.first + 1) %
                                    rmix_threads_data.jobs.size;
    /* Check for a decrease of the size of the thread pools job queue. */
    if (rmix_threads_data.jobs.current == 
        rmix_threads_data.jobs.maximum) {
      /* Compact wrapping job queue of the thread pool. */
      index = (rmix_threads_data.jobs.first +
               rmix_threads_data.jobs.current) %
               rmix_threads_data.jobs.size;
      memcpy(rmix_threads_data.jobs.queue + index,
             rmix_threads_data.jobs.queue +
             rmix_threads_data.jobs.first,
             (rmix_threads_data.jobs.size -
              rmix_threads_data.jobs.first)*
             sizeof(rmix_threads_data.jobs.queue[0]));
      rmix_threads_data.jobs.first = index;
      /* Reallocate the job queue of the thread pool. */
      rmix_threads_data.jobs.size = rmix_threads_data.jobs.maximum;
      if (NULL == (queue =
                   realloc(rmix_threads_data.jobs.queue,
                           rmix_threads_data.jobs.size *
                    sizeof(rmix_threads_data.jobs.queue[0])))) {
        RMIX_LOG((RMIX_WARN(unable to reallocate the job queue)))
        errno = 0;
      } else {
        rmix_threads_data.jobs.queue = queue;
      }
    }
    /* Check for space in job queue. */
    if (rmix_threads_data.jobs.current <
        rmix_threads_data.jobs.maximum) {
      /* Signal dequeued job. */
      if (0 != (error =
                pthread_cond_signal(&rmix_threads_data.jobs.dequeued))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(unable to signal dequeued job)))
        errno = error;
        rmix_syserr();
        /* Check for job cleanup function. */
        if (NULL != cleanup) {
          /* Cleanup queued job. */
          cleanup(argument);
        }
        errno = error;
        pthread_exit(NULL);
      }
    }
    /* Pop unlock thread pool mutex from pthread cleanup stack. */
    pthread_cleanup_pop(0);
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
      /* Check for job cleanup function. */
      if (NULL != cleanup) {
        /* Cleanup queued job. */
        cleanup(argument);
      }
      errno = error;
      pthread_exit(NULL);
    }
#ifdef DEBUG
    /* Check job function. */
    if (NULL == function) {
      errno = EINVAL;
      RMIX_LOG((RMIX_WARN(function is NULL)))
      errno = EINVAL;
      rmix_syserr();
      /* Check for job cleanup function. */
      if (NULL != cleanup) {
        /* Cleanup queued job. */
        cleanup(argument);
      }
    } else {
#endif /* DEBUG */
      /* Execute job function. */
      if (0 != (error = function(cleanup, argument))) {
        errno = error;
        RMIX_LOG((RMIX_WARN(system error in job function)))
        errno = error;
        rmix_syserr();
        /* Check for job cleanup function. */
        if (NULL != cleanup) {
          /* Cleanup queued job. */
          cleanup(argument);
        }
      }
      /* Check for thread cancellation. */
      pthread_testcancel();
#ifdef DEBUG
    }
#endif /* DEBUG */
  }
  /* Pop exit thread from pthread cleanup stack. */
  pthread_cleanup_pop(0);
}

/*
 * The thread pool thread exit cleanup.
 *
 * argument = NULL.
 * return   = NULL.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_exit"
void rmix_threads_exit (void *argument) {
  int          error;
  pthread_t    thread, *threads;
  unsigned int index;
  /* Lock thread pool mutex. */
  if (0 != (error = pthread_mutex_lock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to lock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
    return;
  }
  /* Check for finalized thread pool. */
  if (0 == rmix_threads_data.ready) {
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    return;
  }
  /* Search for current thread in thread pool. */
  thread = pthread_self();
  for (index = 0; index < rmix_threads_data.pool.current; index++) {
    if (0 != pthread_equal(thread, rmix_threads_data.pool.threads[index])) {
      break;
    }
  }
  /* Check if current thread is not in thread pool. */
  if (index == rmix_threads_data.pool.current) {
    errno = ESRCH;
    RMIX_LOG((RMIX_WARN(unable find thread in thread pool)))
    errno = ESRCH;
    rmix_syserr();
    /* Unlock thread pool mutex. */
    if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
      errno = error;
      RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
      errno = error;
      rmix_syserr();
    }
    errno = ESRCH;
    return;
  }
  /* Remove current thread from thread pool. */
  rmix_threads_data.pool.current--;
  memmove(rmix_threads_data.pool.threads + index,
          rmix_threads_data.pool.threads + index + 1,
         (rmix_threads_data.pool.current - index) *
   sizeof(rmix_threads_data.pool.threads[0]));
  /* Check for a decrease of the size of the thread pool. */
  if (rmix_threads_data.pool.current ==
      rmix_threads_data.pool.maximum) {
    /* Reallocate the thread pool. */
    rmix_threads_data.pool.size = rmix_threads_data.pool.maximum;
    if (0 == rmix_threads_data.pool.size) {
      free(rmix_threads_data.pool.threads);
      rmix_threads_data.pool.threads = NULL;
    } else if (NULL == (threads =
                        realloc(rmix_threads_data.pool.threads,
                                rmix_threads_data.pool.size *
                         sizeof(rmix_threads_data.pool.threads[0])))) {
      RMIX_LOG((RMIX_WARN(unable to reallocate the thread pool)))
    } else {
      rmix_threads_data.pool.threads = threads;
    }
  }
  /* Signal thread exit. */
  if (0 != (error = pthread_cond_signal(&rmix_threads_data.pool.exited))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to signal thread exit)))
    errno = error;
    rmix_syserr();
  }
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
  }
}

/*
 * The thread pool mutex unlock cleanup.
 *
 * argument = NULL.
 */
#undef  __FUNC__
#define __FUNC__ "rmix_threads_unlock"
void rmix_threads_unlock (void *argument) {
  int error;
  /* Unlock thread pool mutex. */
  if (0 != (error = pthread_mutex_unlock(&rmix_threads_data.mutex))) {
    errno = error;
    RMIX_LOG((RMIX_WARN(unable to unlock thread pool mutex)))
    errno = error;
    rmix_syserr();
    errno = error;
  }
}

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
