/******************************************************************************
 *
 * Header file for rmix-rpcx stream (XDR data stream) module.
 * Copyright (c) 2002-2006 Oak Ridge National Laboratory.
 *
 * For more information see the following files in the source distribution top-
 * level directory or package data directory (usually /usr/local/share/package):
 *
 * - README    for general package information.
 * - INSTALL   for package install information.
 * - COPYING   for package license information and copying conditions.
 * - AUTHORS   for package authors information.
 * - ChangeLog for package changes information.
 *
 ******************************************************************************/

/* Avoid to include the content of this header file twice. */
#ifndef RMIX_RPCX_STREAM_H
#define RMIX_RPCX_STREAM_H

/******************************************************************************
 *
 * Macros
 *
 ******************************************************************************/

/******************************************************************************
 *
 * Includes
 *
 ******************************************************************************/

/* None at this time. */

/******************************************************************************
 *
 * Data Types
 *
 ******************************************************************************/

/* The RPCX authentication flavor data type. */
typedef enum {
  RMIX_RPCX_STREAM_AUTHFLAV_NONE = 0,
  RMIX_RPCX_STREAM_AUTHFLAV_SYS,
  RMIX_RPCX_STREAM_AUTHFLAV_SHORT,
} rmix_rpcx_stream_authflav_t;

/* The RPCX message type data type. */
typedef enum {
  RMIX_RPCX_STREAM_MSGTYPE_CALL = 0,
  RMIX_RPCX_STREAM_MSGTYPE_REPLY
} rmix_rpcx_stream_msgtype_t;

/* The RPCX reply status data type. */
typedef enum {
  RMIX_RPCX_STREAM_REPLYSTAT_ACCEPTED = 0,
  RMIX_RPCX_STREAM_REPLYSTAT_DENIED
} rmix_rpcx_stream_replystat_t;

/* The RPCX accept status data type. */
typedef enum {
  RMIX_RPCX_STREAM_ACCEPTSTAT_SUCCESS = 0,
  RMIX_RPCX_STREAM_ACCEPTSTAT_PROG_UNAVAIL,
  RMIX_RPCX_STREAM_ACCEPTSTAT_PROG_MISMATCH,
  RMIX_RPCX_STREAM_ACCEPTSTAT_PROC_UNAVAIL,
  RMIX_RPCX_STREAM_ACCEPTSTAT_GARBAGE_ARGS,
  RMIX_RPCX_STREAM_ACCEPTSTAT_SYSTEM_ERR
} rmix_rpcx_stream_acceptstat_t;

/* The RPCX reject status data type. */
typedef enum {
  RMIX_RPCX_STREAM_REJECTSTAT_RPC_MISMATCH = 0,
  RMIX_RPCX_STREAM_REJECTSTAT_AUTH_ERROR,
} rmix_rpcx_stream_rejectstat_t;

/* The RPCX authentication status data type. */
typedef enum {
  RMIX_RPCX_STREAM_AUTHSTAT_OK = 0,
  RMIX_RPCX_STREAM_AUTHSTAT_BADCRED,
  RMIX_RPCX_STREAM_AUTHSTAT_REJECTEDCRED,
  RMIX_RPCX_STREAM_AUTHSTAT_BADVERF,
  RMIX_RPCX_STREAM_AUTHSTAT_REJECTEDVERF,
  RMIX_RPCX_STREAM_AUTHSTAT_TOOWEAK,
  RMIX_RPCX_STREAM_AUTHSTAT_INVALIDRESP,
  RMIX_RPCX_STREAM_AUTHSTAT_FAILED
} rmix_rpcx_stream_authstat_t;

/******************************************************************************
 *
 * Function Prototypes
 *
 ******************************************************************************/

/*
 * Writes a call into an outgoing XDR data stream.
 *
 * ostream = The outgoing XDR data stream.
 * callid  = The call id.
 * prognum = The program number.
 * progver = The program version.
 * procnum = The procedure number.
 * creflav = The authentication credential flavor.
 * cresize = The authentication credential body length.
 * crebody = The authentication credential body.
 * verflav = The authentication verifier flavor.
 * versize = The authentication verifier body length.
 * verbody = The authentication verifier body.
 * sigstr  = The type signature.
 * inary   = The input values array.
 * incnt   = The input values count.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_rpcx_stream_write_call (rmix_ostream_t               *ostream,
                                 unsigned int                  callid,
                                 unsigned int                  prognum,
                                 unsigned int                  progver,
                                 unsigned int                  procnum,
                                 rmix_rpcx_stream_authflav_t   creflav,
                                 unsigned int                  cresize,
                                 const char                   *crebody,
                                 rmix_rpcx_stream_authflav_t   verflav,
                                 unsigned int                  versize,
                                 const char                   *verbody,
                                 const char                   *sigstr,
                                 const void                  **inary,
                                 unsigned int                  incnt);

/*
 * Reads a call without parmeters from an incoming XDR data stream.
 * Writes error messages to an outgoing XDR data stream.
 *
 * istream = The incoming XDR data stream.
 * ostream = The outgoing XDR data stream.
 * callid  = The call id (return).
 * prognum = The program number (return).
 * progver = The program version (return).
 * procnum = The procedure number (return).
 * creflav = The authentication credential flavor (return).
 * cresize = The authentication credential body length (return).
 * crebody = The authentication credential body (return).
 * verflav = The authentication verifier flavor (return).
 * versize = The authentication verifier body length (return).
 * verbody = The authentication verifier body (return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_rpcx_stream_read_call (rmix_istream_t               *istream,
                                rmix_ostream_t               *ostream,
                                unsigned int                 *callid,
                                unsigned int                 *prognum,
                                unsigned int                 *progver,
                                unsigned int                 *procnum,
                                rmix_rpcx_stream_authflav_t  *creflav,
                                unsigned int                 *cresize,
                                char                        **crebody,
                                rmix_rpcx_stream_authflav_t  *verflav,
                                unsigned int                 *versize,
                                char                        **verbody);

/*
 * Creates and reads call parameters from an incoming XDR data stream.
 * Writes error messages to an outgoing XDR data stream.
 *
 * istream = The incoming XDR data stream.
 * ostream = The outgoing XDR data stream.
 * sigstr  = The type signature.
 * inary   = The input values array (return).
 * incnt   = The input values count (return).
 * callid  = The call id (for error messages).
 * verflav = The authentication verifier flavor (for error messages).
 * versize = The authentication verifier body length (for error messages).
 * verbody = The authentication verifier body (for error messages).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 */
int rmix_rpcx_stream_read_call2 (rmix_istream_t                *istream,
                                 rmix_ostream_t                *ostream,
                                 const char                    *sigstr,
                                 void                        ***inary,
                                 unsigned int                  *incnt,
                                 unsigned int                   callid,
                                 rmix_rpcx_stream_authflav_t    verflav,
                                 unsigned int                   versize,
                                 const char                    *verbody);

/*
 * Writes a deny message with an authentication error into an outgoing XDR data
 * stream.
 *
 * ostream  = The outgoing XDR data stream.
 * callid   = The call id.
 * authstat = The authentication status.
 * return   = 0 on success or -1 on any error with errno set appropriately.
 *            ECONNABORTED = Connection closed before data transfer.
 *            EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_write_deny (rmix_ostream_t              *ostream,
                                 unsigned int                 callid,
                                 rmix_rpcx_stream_authstat_t  authstat);

/*
 * Writes an accept message with an error into an outgoing XDR data stream.
 *
 * ostream = The outgoing XDR data stream.
 * callid  = The call id.
 * verflav = The authentication verifier flavor.
 * versize = The authentication verifier body length.
 * verbody = The authentication verifier body.
 * accstat = The accept status.
 * low     = The lowest version supported (if program version mismatch).
 * high    = The highest version supported (if program version mismatch).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_write_error(rmix_ostream_t                *ostream,
                                 unsigned int                   callid,
                                 rmix_rpcx_stream_authflav_t    verflav,
                                 unsigned int                   versize,
                                 const char                    *verbody,
                                 rmix_rpcx_stream_acceptstat_t  accstat,
                                 int                            low,
                                 int                            high);

/*
 * Writes a reply message w/o procedure return into an outgoing XDR data stream.
 *
 * ostream = The outgoing XDR data stream.
 * callid  = The call id.
 * verflav = The authentication verifier flavor.
 * versize = The authentication verifier body length.
 * verbody = The authentication verifier body.
 * sigstr  = The type signature.
 * outary  = The output values array (also return).
 * outcnt  = The output values array size (also return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_write_reply(rmix_ostream_t              *ostream,
                                 unsigned int                 callid,
                                 rmix_rpcx_stream_authflav_t  verflav,
                                 unsigned int                 versize,
                                 const char                  *verbody);

/*
 * Writes and destroys the procedure return of a reply message.
 *
 * ostream = The outgoing XDR data stream.
 * sigstr  = The type signature.
 * outary  = The output values array (also return).
 * outcnt  = The output values array size (also return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_write_reply2(rmix_ostream_t   *ostream,
                                  const char       *sigstr,
                                  void           ***outary,
                                  unsigned int     *outcnt);

/*
 * Reads a reply message w/o procedure return from an incoming XDR data stream.
 *
 * istream = The incoming XDR data stream.
 * callid  = The call id.
 * verflav = The authentication verifier flavor (return).
 * versize = The authentication verifier body length (return).
 * verbody = The authentication verifier body (return).
 * low     = The lowest version supported (return if RPC/progver mismatch).
 * high    = The highest version supported (return if RPC/progver mismatch).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 *           ESRCH        = Remote prognum unavailable.
 *           ENOTSUP      = Remote progver mismatch.
 *           EOPNOTSUPP   = Remote procnum unavailable.
 *           EBADMSG      = Remote garbled args.
 *           ENOSYS       = Remote system error.
 *           ENOPROTOOPT  = Remote RPC version mismatch.
 *           EACCES       = Remote authentication error.
 *           EPROTO       = Protocol error.
 */
int rmix_rpcx_stream_read_reply (rmix_istream_t               *istream,
                                 unsigned int                  callid,
                                 rmix_rpcx_stream_authflav_t  *verflav,
                                 unsigned int                 *versize,
                                 char                        **verbody,
                                 int                          *low,
                                 int                          *high);

/*
 * Reads the procedure return of a reply message. String and variable array
 * values are allocated.
 *
 * istream = The incoming XDR data stream.
 * sigstr  = The type signature.
 * outary  = The output values array (also return).
 * outcnt  = The output values array size.
 * low     = The lowest version supported (return if RPC/progver mismatch).
 * high    = The highest version supported (return if RPC/progver mismatch).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_read_reply2 (rmix_istream_t  *istream,
                                  const char      *sigstr,
                                  void           **outary,
                                  unsigned int     outcnt);

/*
 * Writes a values array into an outgoing XDR data stream.
 *
 * ostream = The outgoing XDR data stream.
 * valary  = The values array.
 * valcnt  = The values array size.
 * sigstr  = The type signature.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_valary_write (rmix_ostream_t  *ostream,
                                   const void     **valary,
                                   unsigned int     valcnt,
                                   const char      *sigstr);

/*
 * Writes a values array into an outgoing XDR data stream and destroys it.
 *
 * istream = The incoming XDR data stream.
 * valary  = The values array.
 * valcnt  = The values array size (return).
 * sigstr  = The type signature.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_valary_write2 (rmix_ostream_t   *ostream,
                                    void           ***valary,
                                    unsigned int     *valcnt,
                                    const char       *sigstr);

/*
 * Reads a values array from an incoming XDR data stream. String and variable
 * array values are allocated.
 *
 * istream = The incoming XDR data stream.
 * valary  = The values array.
 * valcnt  = The values array size.
 * sigstr  = The type signature.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_valary_read (rmix_istream_t  *istream,
                                  void           **valary,
                                  unsigned int     valcnt,
                                  const char      *sigstr);

/*                 
 * Creates and reads a values array from an incoming XDR data stream.
 *
 * istream = The incoming XDR data stream.
 * valary  = The values array.
 * valcnt  = The values array size (return).
 * sigstr  = The type signature.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_valary_read2 (rmix_istream_t   *istream,
                                   void           ***valary,
                                   unsigned int     *valcnt,
                                   const char       *sigstr);

/*
 * Writes a value into an outgoing XDR data stream and advances the values array
 * and signature indices.
 *
 * ostream = The outgoing XDR data stream.
 * valary  = The values array.
 * validx  = The values index (also return).
 * sigstr  = The type signature.
 * sigidx  = The type signature index (also return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_value_write (rmix_ostream_t  *ostream,
                                  const void     **valary,
                                  unsigned int    *validx,
                                  const char      *sigstr,
                                  unsigned int    *sigidx);

/*
 * Reads a value from an incoming XDR data stream and advances the values array
 * and signature indices. String and variable array values are allocated.
 *
 * istream = The incoming XDR data stream.
 * valary  = The values array.
 * validx  = The values index (also return).
 * sigstr  = The type signature.
 * sigidx  = The type signature index (also return).
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_value_read (rmix_istream_t  *istream,
                                 void           **valary,
                                 unsigned int    *validx,
                                 const char      *sigstr,
                                 unsigned int    *sigidx);

/*
 * Writes values into an outgoing XDR data stream.
 *
 * ostream = The outgoing XDR data stream.
 * count   = The number of values to write.
 * length  = The byte length of a single value.
 * values  = The values buffer to write from.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_write (rmix_ostream_t *ostream,
                            unsigned int    count,
                            unsigned int    length,
                            const void     *values);

/*
 * Reads values from an incoming XDR data stream.
 *
 * istream = The incoming XDR data stream.
 * count   = The number of values to read.
 * length  = The byte length of a single value.
 * values  = The values buffer to read into.
 * return  = 0 on success or -1 on any error with errno set appropriately.
 *           ECONNABORTED = Connection closed before data transfer.
 *           EPIPE        = Connection closed during data transfer.
 */
int rmix_rpcx_stream_read (rmix_istream_t *istream,
                           unsigned int    count,
                           unsigned int    length,
                           void           *values);

/******************************************************************************
 *
 * Data
 *
 ******************************************************************************/

/* None at this time. */

#endif /* RMIX_RPCX_STREAM_H */

/******************************************************************************
 *
 * END OF FILE
 *
 ******************************************************************************/
